# test mod_md acme terms-of-service handling

import re

from TestEnv import TestEnv


def setup_module(module):
    print("setup_module: %s" % module.__name__)
    TestEnv.init()
        

def teardown_module(module):
    print("teardown_module: %s" % module.__name__)


def md_name(md):
    return md['name']


class TestStore:

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        TestEnv.purge_store()
 
    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    # verify expected binary version
    def test_000_001(self):
        run = TestEnv.run([TestEnv.A2MD, "-V"])
        m = re.match("version: %s(-git)?$" %
                     TestEnv.config.get('global', 'a2md_version'), run['stdout'])
        assert m

    # verify that store is clean
    def test_000_002(self):
        run = TestEnv.run(["find", TestEnv.STORE_DIR])
        assert re.match(TestEnv.STORE_DIR, run['stdout'])

    # test case: add a single dns managed domain
    def test_000_100(self):
        dns = "greenbytes.de"
        TestEnv.check_json_contains(
            TestEnv.a2md(["store", "add", dns])['jout']['output'][0],
            {
                "name": dns,
                "domains": [dns],
                "contacts": [],
                "ca": {
                    "url": TestEnv.ACME_URL,
                    "proto": "ACME"
                },
                "state": 0
            })

    # test case: add > 1 dns managed domain
    def test_000_101(self):
        dns = ["greenbytes2.de", "www.greenbytes2.de", "mail.greenbytes2.de"]
        TestEnv.check_json_contains(
            TestEnv.a2md(["store", "add"] + dns)['jout']['output'][0],
            {
                "name": dns[0],
                "domains": dns,
                "contacts": [],
                "ca": {
                    "url": TestEnv.ACME_URL,
                    "proto": "ACME"
                },
                "state": 0
            })

    # test case: add second managed domain
    def test_000_102(self):
        dns1 = ["test000-102.com", "test000-102a.com", "test000-102b.com"]
        assert TestEnv.a2md(["store", "add"] + dns1)['rv'] == 0
        #
        # add second managed domain
        dns2 = ["greenbytes2.de", "www.greenbytes2.de", "mail.greenbytes2.de"]
        jout = TestEnv.a2md(["store", "add"] + dns2)['jout']
        # assert: output covers only changed md
        assert len(jout['output']) == 1
        TestEnv.check_json_contains(jout['output'][0], {
            "name": dns2[0],
            "domains": dns2,
            "contacts": [],
            "ca": {
                "url": TestEnv.ACME_URL,
                "proto": "ACME"
            },
            "state": 0
        })

    # test case: add existing domain 
    def test_000_103(self):
        dns = "greenbytes.de"
        assert TestEnv.a2md(["store", "add", dns])['rv'] == 0
        # add same domain again
        assert TestEnv.a2md(["store", "add", dns])['rv'] == 1

    # test case: add without CA URL
    def test_000_104(self):
        dns = "greenbytes.de"
        args = [TestEnv.A2MD, "-d", TestEnv.STORE_DIR, "-j", "store", "add", dns]
        jout = TestEnv.run(args)['jout']
        assert len(jout['output']) == 1
        TestEnv.check_json_contains(jout['output'][0], {
            "name": dns,
            "domains": [dns],
            "contacts": [],
            "ca": {
                "proto": "ACME"
            },
            "state": 0
        })

    # test case: list empty store
    def test_000_200(self):
        assert TestEnv.a2md(["store", "list"])['jout'] == TestEnv.EMPTY_JOUT

    # test case: list two managed domains
    def test_000_201(self):
        domains = [ 
            ["test000-201.com", "test000-201a.com", "test000-201b.com"],
            ["greenbytes2.de", "www.greenbytes2.de", "mail.greenbytes2.de"]
        ]
        for dns in domains:
            assert TestEnv.a2md(["store", "add"] + dns)['rv'] == 0
        #
        # list all store content
        jout = TestEnv.a2md(["store", "list"])['jout']
        assert len(jout['output']) == len(domains)
        domains.reverse()
        jout['output'] = sorted(jout['output'], key=md_name)
        for i in range(0, len(jout['output'])):
            TestEnv.check_json_contains(jout['output'][i], {
                "name": domains[i][0],
                "domains": domains[i],
                "contacts": [],
                "ca": {
                    "url": TestEnv.ACME_URL,
                    "proto": "ACME"
                },
                "state": 0
            })

    # test case: remove managed domain
    def test_000_300(self):
        dns = "test000-300.com"
        assert TestEnv.a2md(["store", "add", dns])['rv'] == 0
        assert TestEnv.a2md(["store", "remove", dns])['jout'] == TestEnv.EMPTY_JOUT
        assert TestEnv.a2md(["store", "list"])['jout'] == TestEnv.EMPTY_JOUT

    # test case: remove from list of managed domains 
    def test_000_301(self):
        dns1 = ["test000-301.com", "test000-301a.com", "test000-301b.com"]
        assert TestEnv.a2md(["store", "add"] + dns1)['rv'] == 0
        #
        dns2 = ["greenbytes2.de", "www.greenbytes2.de", "mail.greenbytes2.de"]
        jout1 = TestEnv.a2md(["store", "add"] + dns2)['jout']
        # remove managed domain
        assert TestEnv.a2md(["store", "remove", "test000-301.com"])['jout'] == TestEnv.EMPTY_JOUT
        # list store content
        assert TestEnv.a2md(["store", "list"])['jout'] == jout1

    # test case: remove nonexisting managed domain
    def test_000_302(self):
        dns1 = "test000-302.com"
        run = TestEnv.a2md(["store", "remove", dns1])
        assert run['rv'] == 1
        assert run['jout'] == { 
            'status': 2, 'description': 'No such file or directory', 'output': []
        }

    # test case: force remove nonexisting managed domain
    def test_000_303(self):
        dns1 = "test000-303.com"
        assert TestEnv.a2md(["store", "remove", "-f", dns1])['jout'] == TestEnv.EMPTY_JOUT

    # test case: null change
    def test_000_400(self):
        dns = "test000-400.com"
        run1 = TestEnv.a2md(["store", "add", dns])
        assert TestEnv.a2md(["store", "update", dns])['jout'] == run1['jout']

    # test case: add dns to managed domain
    def test_000_401(self):
        dns1 = "test000-401.com"
        TestEnv.a2md(["store", "add", dns1])
        dns2 = "test-101.com"
        args = ["store", "update", dns1, "domains", dns1, dns2]
        assert TestEnv.a2md(args)['jout']['output'][0]['domains'] == [dns1, dns2]

    # test case: change CA URL
    def test_000_402(self):
        dns = "test000-402.com"
        args = ["store", "add", dns]
        assert TestEnv.a2md(args)['jout']['output'][0]['ca']['url'] == TestEnv.ACME_URL
        nurl = "https://foo.com/"
        args = [TestEnv.A2MD, "-a", nurl, "-d", TestEnv.STORE_DIR, "-j", "store", "update", dns]
        assert TestEnv.run(args)['jout']['output'][0]['ca']['url'] == nurl

    # test case: update nonexisting managed domain
    def test_000_403(self):
        dns = "test000-403.com"
        assert TestEnv.a2md(["store", "update", dns])['rv'] == 1

    # test case: update domains, throw away md name
    def test_000_404(self):
        dns1 = "test000-404.com"
        dns2 = "greenbytes.com"
        args = ["store", "add", dns1]
        assert TestEnv.a2md(args)['jout']['output'][0]['domains'] == [dns1]
        # override domains list
        args = ["store", "update", dns1, "domains", dns2]
        assert TestEnv.a2md(args)['jout']['output'][0]['domains'] == [dns2]

    # test case: update domains with empty dns list
    def test_000_405(self):
        dns1 = "test000-405.com"
        assert TestEnv.a2md(["store", "add", dns1])['rv'] == 0
        assert TestEnv.a2md(["store", "update", dns1, "domains"])['rv'] == 1



# test mod_md stapling support

import os
import time
import pytest

from TestEnv import TestEnv
from TestHttpdConf import HttpdConf


@pytest.mark.skipif(TestEnv.ACME_LACKS_OCSP, reason="no OCSP responder")
class TestStapling:
    domain = None
    mdA = None
    mdB = None

    @classmethod
    def setup_class(cls):
        print("setup_class:%s" % cls.__name__)
        TestEnv.init()
        TestEnv.clear_store()
        TestEnv.check_acme()
        cls.domain = TestEnv.get_class_domain(cls)
        cls.mdA = "a-" + cls.domain
        cls.mdB = "b-" + cls.domain
        cls.configure_httpd([cls.mdA, cls.mdB]).install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([cls.mdA, cls.mdB])
        TestEnv.check_md_complete(cls.mdA)
        TestEnv.check_md_complete(cls.mdB)

    @classmethod
    def teardown_class(cls):
        print("teardown_class:%s" % cls.__name__)
        assert TestEnv.apache_stop() == 0
    
    @classmethod
    def configure_httpd(cls, domains=None, add_lines="", ssl_stapling=False):
        if not isinstance(domains, list):
            domains = [domains] if domains else []
        conf = HttpdConf()
        conf.add_admin("admin@" + cls.domain)
        if ssl_stapling:
            conf.add_line("""
                LogLevel ssl:trace2
                SSLUseStapling On
                SSLStaplingCache \"shmcb:logs/ssl_stapling(32768)\"
                """)
        conf.add_line(add_lines)
        for domain in domains:
            conf.add_md([domain])
            conf.add_vhost(domain)
        return conf
    
    # MD with stapling on/off and mod_ssl stapling off
    # expect to only see stapling response when MD stapling is on
    def test_801_001(self):
        md = TestStapling.mdA
        TestStapling.configure_httpd(md).install()
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.get_ocsp_status(md)
        assert stat['ocsp'] == "no response sent" 
        stat = TestEnv.get_md_status(md)
        assert not stat["stapling"]
        #
        # turn stapling on, wait for it to appear in connections
        TestStapling.configure_httpd(md, """
            MDStapling on
            LogLevel md:trace5
            """).install()
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.await_ocsp_status(md)
        assert stat['ocsp'] == "successful (0x0)" 
        assert stat['verify'] == "0 (ok)"
        stat = TestEnv.get_md_status(md)
        assert stat["stapling"]
        pkey = 'rsa'
        assert stat["cert"][pkey]["ocsp"]["status"] == "good"
        assert stat["cert"][pkey]["ocsp"]["valid"]
        #
        # turn stapling off (explicitly) again, should disappear
        TestStapling.configure_httpd(md, "MDStapling off").install()
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.get_ocsp_status(md)
        assert stat['ocsp'] == "no response sent" 
        stat = TestEnv.get_md_status(md)
        assert not stat["stapling"]
        
    # MD with stapling on/off and mod_ssl stapling on
    # expect to see stapling response in all cases
    def test_801_002(self):
        md = TestStapling.mdA
        TestStapling.configure_httpd(md, ssl_stapling=True).install()
        assert TestEnv.apache_stop() == 0
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.get_ocsp_status(md)
        assert stat['ocsp'] == "successful (0x0)" 
        stat = TestEnv.get_md_status(md)
        assert not stat["stapling"]
        #
        # turn stapling on, wait for it to appear in connections
        TestStapling.configure_httpd(md, "MDStapling on", ssl_stapling=True).install()
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.await_ocsp_status(md)
        assert stat['ocsp'] == "successful (0x0)" 
        assert stat['verify'] == "0 (ok)"
        stat = TestEnv.get_md_status(md)
        assert stat["stapling"]
        pkey = 'rsa'
        assert stat["cert"][pkey]["ocsp"]["status"] == "good"
        assert stat["cert"][pkey]["ocsp"]["valid"]
        #
        # turn stapling off (explicitly) again, should disappear
        TestStapling.configure_httpd(md, "MDStapling off", ssl_stapling=True).install()
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.get_ocsp_status(md)
        assert stat['ocsp'] == "successful (0x0)" 
        stat = TestEnv.get_md_status(md)
        assert not stat["stapling"]
        
    # 2 MDs, one with md stapling on, one with default (off)
    def test_801_003(self):
        md_a = TestStapling.mdA
        md_b = TestStapling.mdB
        conf = TestStapling.configure_httpd()
        conf.add_line("""
            <MDomain %s>
                MDStapling on
            </MDomain>
            <MDomain %s>
            </MDomain>
            """ % (md_a, md_b))
        conf.add_vhost(md_a)
        conf.add_vhost(md_b)
        conf.install()
        assert TestEnv.apache_stop() == 0
        assert TestEnv.apache_restart() == 0
        # mdA has stapling
        stat = TestEnv.await_ocsp_status(md_a)
        assert stat['ocsp'] == "successful (0x0)" 
        assert stat['verify'] == "0 (ok)"
        stat = TestEnv.get_md_status(md_a)
        assert stat["stapling"]
        pkey = 'rsa'
        assert stat["cert"][pkey]["ocsp"]["status"] == "good"
        assert stat["cert"][pkey]["ocsp"]["valid"]
        # mdB has no stapling
        stat = TestEnv.get_ocsp_status(md_b)
        assert stat['ocsp'] == "no response sent" 
        stat = TestEnv.get_md_status(md_b)
        assert not stat["stapling"]

    # 2 MDs, md stapling on+off, ssl stapling on
    def test_801_004(self):
        md_a = TestStapling.mdA
        md_b = TestStapling.mdB
        conf = TestStapling.configure_httpd(ssl_stapling=True)
        conf.add_line("""
            <MDomain %s>
                MDStapling on
            </MDomain>
            <MDomain %s>
            </MDomain>
            """ % (md_a, md_b))
        conf.add_vhost(md_a)
        conf.add_vhost(md_b)
        conf.install()
        assert TestEnv.apache_stop() == 0
        assert TestEnv.apache_restart() == 0
        # mdA has stapling
        stat = TestEnv.await_ocsp_status(md_a)
        assert stat['ocsp'] == "successful (0x0)" 
        assert stat['verify'] == "0 (ok)"
        stat = TestEnv.get_md_status(md_a)
        assert stat["stapling"]
        pkey = 'rsa'
        assert stat["cert"][pkey]["ocsp"]["status"] == "good"
        assert stat["cert"][pkey]["ocsp"]["valid"]
        # mdB has no md stapling, but mod_ssl kicks in
        stat = TestEnv.get_ocsp_status(md_b)
        assert stat['ocsp'] == "successful (0x0)" 
        stat = TestEnv.get_md_status(md_b)
        assert not stat["stapling"]

    # MD, check that restart leaves response unchanged, reconfigure keep interval, 
    # should remove the file on restart and get a new one
    def test_801_005(self):
        # TODO: mod_watchdog seems to have problems sometimes with fast restarts
        # stopping first works.
        assert TestEnv.apache_stop() == 0
        # turn stapling on, wait for it to appear in connections
        md = TestStapling.mdA
        TestStapling.configure_httpd(md, "MDStapling on").install()
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.await_ocsp_status(md)
        assert stat['ocsp'] == "successful (0x0)" 
        assert stat['verify'] == "0 (ok)"
        # fine the file where the ocsp response is stored
        dirpath = os.path.join(TestEnv.STORE_DIR, 'ocsp', md)
        files = os.listdir(dirpath)
        ocsp_file = None
        for name in files:
            if name.startswith("ocsp-"):
                ocsp_file = os.path.join(dirpath, name)
        assert ocsp_file
        mtime1 = os.path.getmtime(ocsp_file)
        # wait a sec, restart and check that file does not change
        time.sleep(1)
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.await_ocsp_status(md)
        assert stat['ocsp'] == "successful (0x0)" 
        mtime2 = os.path.getmtime(ocsp_file)
        assert mtime1 == mtime2
        # configure a keep time of 1 second, restart, the file is gone
        # (which is a side effec that we load it before the cleanup removes it.
        #  since it was valid, no new one needed fetching
        TestStapling.configure_httpd(md, """
            MDStapling on
            MDStaplingKeepResponse 1s
            """).install()
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.await_ocsp_status(md)
        assert stat['ocsp'] == "successful (0x0)"
        assert not os.path.exists(ocsp_file)
        # if we restart again, a new file needs to appear
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.await_ocsp_status(md)
        assert stat['ocsp'] == "successful (0x0)"
        mtime3 = os.path.getmtime(ocsp_file)
        assert mtime1 != mtime3

    # MD, check that stapling renew window works. Set a large window
    # that causes response to be retrieved all the time.
    def test_801_006(self):
        assert TestEnv.apache_stop() == 0
        # turn stapling on, wait for it to appear in connections
        md = TestStapling.mdA
        TestStapling.configure_httpd(md, "MDStapling on").install()
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.await_ocsp_status(md)
        assert stat['ocsp'] == "successful (0x0)" 
        assert stat['verify'] == "0 (ok)"
        # fine the file where the ocsp response is stored
        dirpath = os.path.join(TestEnv.STORE_DIR, 'ocsp', md)
        files = os.listdir(dirpath)
        ocsp_file = None
        for name in files:
            if name.startswith("ocsp-"):
                ocsp_file = os.path.join(dirpath, name)
        assert ocsp_file
        mtime1 = os.path.getmtime(ocsp_file)
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.await_ocsp_status(md)
        assert stat['ocsp'] == "successful (0x0)" 
        # wait a sec, restart and check that file does not change
        time.sleep(1)
        mtime2 = os.path.getmtime(ocsp_file)
        assert mtime1 == mtime2
        # configure a renew window of 10 days, restart, larger than any life time.
        TestStapling.configure_httpd(md, """
            MDStapling on
            MDStaplingRenewWindow 10d
            """).install()
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.await_ocsp_status(md)
        assert stat['ocsp'] == "successful (0x0)"
        # wait a sec, restart and check that file does change
        time.sleep(1)
        mtime3 = os.path.getmtime(ocsp_file)
        assert mtime1 != mtime3

    # MD, make a MDomain with static files, check that stapling works
    def test_801_007(self):
        assert TestEnv.apache_stop() == 0
        # turn stapling on, wait for it to appear in connections
        md = TestStapling.mdA
        conf = TestStapling.configure_httpd()
        conf.add_line("""
            <MDomain %s>
                MDCertificateKeyFile %s
                MDCertificateFile %s
                MDStapling on
            </MDomain>
            """ % (md, TestEnv.store_domain_file(md, 'privkey.pem'),
                   TestEnv.store_domain_file(md, 'pubcert.pem')))
        conf.add_vhost(md)
        conf.install()
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.await_ocsp_status(md)
        assert stat['ocsp'] == "successful (0x0)" 
        assert stat['verify'] == "0 (ok)"
        # fine the file where the ocsp response is stored
        dirpath = os.path.join(TestEnv.STORE_DIR, 'ocsp', md)
        files = os.listdir(dirpath)
        ocsp_file = None
        for name in files:
            if name.startswith("ocsp-"):
                ocsp_file = os.path.join(dirpath, name)
        assert ocsp_file

    # Use certificate files in old skool mod_ssl set, check that stapling works
    def test_801_008(self):
        assert TestEnv.apache_stop() == 0
        # turn stapling on, wait for it to appear in connections
        md = TestStapling.mdA
        conf = TestStapling.configure_httpd()
        conf.add_line("MDStapling on")
        conf.start_vhost(md)
        conf.add_line("""
            SSLCertificateKeyFile %s
            SSLCertificateFile %s
            """ % (TestEnv.store_domain_file(md, 'privkey.pem'),
                   TestEnv.store_domain_file(md, 'pubcert.pem')))
        conf.end_vhost()
        conf.install()
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.await_ocsp_status(md)
        assert stat['ocsp'] == "successful (0x0)" 
        assert stat['verify'] == "0 (ok)"
        # fine the file where the ocsp response is stored
        dirpath = os.path.join(TestEnv.STORE_DIR, 'ocsp', 'other')
        files = os.listdir(dirpath)
        ocsp_file = None
        for name in files:
            if name.startswith("ocsp-"):
                ocsp_file = os.path.join(dirpath, name)
        assert ocsp_file

    # Turn on stapling for a certificate without OCSP responder and issuer
    # (certificates without issuer prevent mod_ssl asking around for stapling)
    def test_801_009(self):
        assert TestEnv.apache_stop() == 0
        md = TestStapling.mdA
        domains = [md]
        testpath = os.path.join(TestEnv.GEN_DIR, 'test_801_009')
        # cert that is 30 more days valid
        TestEnv.create_self_signed_cert(domains, {"notBefore": -60, "notAfter": 30},
                                        serial=801009, path=testpath)
        cert_file = os.path.join(testpath, 'pubcert.pem')
        pkey_file = os.path.join(testpath, 'privkey.pem')
        assert os.path.exists(cert_file)
        assert os.path.exists(pkey_file)
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.start_md(domains)
        conf.add_line("MDCertificateFile %s" % cert_file)
        conf.add_line("MDCertificateKeyFile %s" % pkey_file)
        conf.add_line("MDStapling on")
        conf.end_md()
        conf.add_vhost(md)
        conf.install()
        assert TestEnv.apache_restart() == 0
        time.sleep(1)
        stat = TestEnv.get_ocsp_status(md)
        assert stat['ocsp'] == "no response sent" 

    # Turn on stapling for an MDomain not used in any virtualhost
    # There was a crash in server-status in this case
    def test_801_010(self):
        assert TestEnv.apache_stop() == 0
        TestEnv.clear_ocsp_store()
        md = TestStapling.mdA
        domains = [md]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.start_md(domains)
        conf.add_line("MDStapling on")
        conf.end_md()
        conf.install()
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.get_server_status()
        assert stat



###################################################################################################
# httpd test configuration generator
#
# (c) 2019 greenbytes GmbH
###################################################################################################

import os
from shutil import copyfile

from TestEnv import TestEnv


class HttpdConf(object):
    # Utility class for creating Apache httpd test configurations

    def __init__(self, name="test.conf", local_ca=True, text=None, std_vhosts=True, proxy=False):
        self.path = os.path.join(TestEnv.GEN_DIR, name)
        if os.path.isfile(self.path):
            os.remove(self.path)
        if not text:
            text = """
LogLevel md:trace2
LogLevel ssl:debug
                
                """
            
        if local_ca:
            text = """
MDCertificateAuthority %s
MDCertificateAgreement accepted
MDCACertificateFile %s/test-ca.pem
                
%s
""" % (TestEnv.ACME_URL, TestEnv.WEBROOT, text)

        if std_vhosts:
            text = """
Listen %s
Listen %s

MDPortMap 80:%s 443:%s
                
include "conf/std_vhosts.conf"
                
%s
""" % (TestEnv.HTTP_PORT, TestEnv.HTTPS_PORT, TestEnv.HTTP_PORT, TestEnv.HTTPS_PORT, text)

        if proxy:
            text = """
include "conf/proxy.conf"
                
%s
""" % text
        open(self.path, "a").write(text)

    def clear(self):
        if os.path.isfile(self.path):
            os.remove(self.path)

    def _add_line(self, line):
        open(self.path, "a").write(line + "\n")

    def add_line(self, line):
        self._add_line(line)

    def add_drive_mode(self, mode):
        self._add_line("  MDRenewMode %s\n" % mode)

    def add_renew_window(self, window):
        self._add_line("  MDRenewWindow %s\n" % window)

    def add_private_key(self, key_type, key_params):
        self._add_line("  MDPrivateKeys %s %s\n" % (key_type, " ".join(map(lambda p: str(p), key_params))))

    def add_admin(self, email):
        self._add_line("  ServerAdmin mailto:%s\n\n" % email)

    def add_md(self, domains):
        self._add_line("  MDomain %s\n\n" % " ".join(domains))

    def start_md(self, domains):
        self._add_line("  <MDomain %s>\n" % " ".join(domains))
        
    def start_md2(self, domains):
        self._add_line("  <MDomainSet %s>\n" % " ".join(domains))

    def end_md(self):
        self._add_line("  </MDomain>\n")

    def end_md2(self):
        self._add_line("  </MDomainSet>\n")

    def add_must_staple(self, mode):
        self._add_line("  MDMustStaple %s\n" % mode)

    def add_ca_challenges(self, type_list):
        self._add_line("  MDCAChallenges %s\n" % " ".join(type_list))

    def add_http_proxy(self, url):
        self._add_line("  MDHttpProxy %s\n" % url)

    def add_require_ssl(self, mode):
        self._add_line("  MDRequireHttps %s\n" % mode)

    def add_notify_cmd(self, cmd):
        self._add_line("  MDNotifyCmd %s\n" % cmd)

    def add_message_cmd(self, cmd):
        self._add_line("  MDMessageCmd %s\n" % cmd)

    def add_dns01_cmd(self, cmd):
        self._add_line("  MDChallengeDns01 %s\n" % cmd)

    def add_vhost(self, domains, port=None, doc_root="htdocs"):
        self.start_vhost(domains, port=port, doc_root=doc_root)
        self.end_vhost()

    def start_vhost(self, domains, port=None, doc_root="htdocs"):
        if not isinstance(domains, list):
            domains = [domains]
        if not port:
            port = TestEnv.HTTPS_PORT 
        f = open(self.path, "a") 
        f.write("<VirtualHost *:%s>\n" % port)
        f.write("    ServerName %s\n" % domains[0])
        for alias in domains[1:]:
            f.write("    ServerAlias %s\n" % alias)
        f.write("    DocumentRoot %s\n\n" % doc_root)
        if TestEnv.HTTPS_PORT == port:
            f.write("    SSLEngine on\n")
                  
    def end_vhost(self):
        self._add_line("</VirtualHost>\n\n")

    def install(self):
        copyfile(self.path, TestEnv.APACHE_TEST_CONF)



# test mod_md basic configurations

import re
import pytest

from configparser import ConfigParser
from TestEnv import TestEnv
from TestHttpdConf import HttpdConf

config = ConfigParser()
config.read('test.ini')
PREFIX = config.get('global', 'prefix')


def setup_module(module):
    print("setup_module    module:%s" % module.__name__)
    TestEnv.init()
    TestEnv.clear_store()
    

def teardown_module(module):
    print("teardown_module module:%s" % module.__name__)
    TestEnv.apache_stop()


class TestConf:

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        TestEnv.httpd_error_log_clear()

    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    # test case: just one MDomain definition
    def test_300_001(self):
        HttpdConf(text="""
            MDomain not-forbidden.org www.not-forbidden.org mail.not-forbidden.org
            """).install()
        assert TestEnv.apache_restart() == 0

    # test case: two MDomain definitions, non-overlapping
    def test_300_002(self):
        HttpdConf(text="""
            MDomain not-forbidden.org www.not-forbidden.org mail.not-forbidden.org
            MDomain example2.org www.example2.org mail.example2.org
            """).install()
        assert TestEnv.apache_restart() == 0

    # test case: two MDomain definitions, exactly the same
    def test_300_003(self):
        assert TestEnv.apache_stop() == 0
        HttpdConf(text="""
            MDomain not-forbidden.org www.not-forbidden.org mail.not-forbidden.org test3.not-forbidden.org
            MDomain not-forbidden.org www.not-forbidden.org mail.not-forbidden.org test3.not-forbidden.org
            """).install()
        assert TestEnv.apache_fail() == 0
        
    # test case: two MDomain definitions, overlapping
    def test_300_004(self):
        assert TestEnv.apache_stop() == 0
        HttpdConf(text="""
            MDomain not-forbidden.org www.not-forbidden.org mail.not-forbidden.org test3.not-forbidden.org
            MDomain example2.org test3.not-forbidden.org www.example2.org mail.example2.org
            """).install()
        assert TestEnv.apache_fail() == 0

    # test case: two MDomains, one inside a virtual host
    def test_300_005(self):
        HttpdConf(text="""
            MDomain not-forbidden.org www.not-forbidden.org mail.not-forbidden.org test3.not-forbidden.org
            <VirtualHost *:12346>
                MDomain example2.org www.example2.org www.example3.org
            </VirtualHost>
            """).install()
        assert TestEnv.apache_restart() == 0

    # test case: two MDomains, one correct vhost name
    def test_300_006(self):
        HttpdConf(text="""
            MDomain not-forbidden.org www.not-forbidden.org mail.not-forbidden.org test3.not-forbidden.org
            <VirtualHost *:12346>
                ServerName example2.org
                MDomain example2.org www.example2.org www.example3.org
            </VirtualHost>
            """).install()
        assert TestEnv.apache_restart() == 0

    # test case: two MDomains, two correct vhost names
    def test_300_007(self):
        HttpdConf(text="""
            MDomain not-forbidden.org www.not-forbidden.org mail.not-forbidden.org test3.not-forbidden.org
            <VirtualHost *:12346>
                ServerName example2.org
                MDomain example2.org www.example2.org www.example3.org
            </VirtualHost>
            <VirtualHost *:12346>
                ServerName www.example2.org
            </VirtualHost>
            """).install()
        assert TestEnv.apache_restart() == 0

    # test case: two MDomains, overlapping vhosts
    def test_300_008(self):
        HttpdConf(text="""
            MDomain not-forbidden.org www.not-forbidden.org mail.not-forbidden.org test3.not-forbidden.org
            <VirtualHost *:12346>
                ServerName example2.org
                ServerAlias www.example3.org
                MDomain example2.org www.example2.org www.example3.org
            </VirtualHost>

            <VirtualHost *:12346>
                ServerName www.example2.org
                ServerAlias example2.org
            </VirtualHost>
            """).install()
        assert TestEnv.apache_restart() == 0

    # test case: vhosts with overlapping MDs
    def test_300_009(self):
        assert TestEnv.apache_stop() == 0
        HttpdConf(text="""
            ServerAdmin admin@not-forbidden.org
            MDMembers manual
            MDomain not-forbidden.org www.not-forbidden.org mail.not-forbidden.org test3.not-forbidden.org
            MDomain example2.org www.example2.org www.example3.org

            <VirtualHost *:12346>
                ServerName example2.org
                ServerAlias www.example3.org
                SSLEngine on
            </VirtualHost>

            <VirtualHost *:12346>
                ServerName www.example2.org
                ServerAlias example2.org
                SSLEngine on
            </VirtualHost>

            <VirtualHost *:12346>
                ServerName not-forbidden.org
                ServerAlias example2.org
                SSLEngine on
            </VirtualHost>
            """).install()
        assert TestEnv.apache_fail() == 0

    # test case: MDomain, vhost with matching ServerAlias
    def test_300_010(self):
        HttpdConf(text="""
            MDomain not-forbidden.org www.not-forbidden.org mail.not-forbidden.org test3.not-forbidden.org

            <VirtualHost *:12346>
                ServerName not-forbidden.org
                ServerAlias test3.not-forbidden.org
            </VirtualHost>
            """).install()
        assert TestEnv.apache_restart() == 0
        assert (0, 0) == TestEnv.httpd_error_log_count()

    # test case: MDomain, misses one ServerAlias
    def test_300_011(self):
        HttpdConf(text="""
            MDomain not-forbidden.org manual www.not-forbidden.org mail.not-forbidden.org test3.not-forbidden.org

            <VirtualHost *:%s>
                ServerName not-forbidden.org
                ServerAlias test3.not-forbidden.org
                ServerAlias test4.not-forbidden.org
                SSLEngine on
            </VirtualHost>
            """ % TestEnv.HTTPS_PORT).install()
        assert TestEnv.apache_fail() == 0
        assert (1, 0) == TestEnv.httpd_error_log_count()

    # test case: MDomain, misses one ServerAlias, but auto add enabled
    def test_300_011b(self):
        assert TestEnv.apache_stop() == 0
        HttpdConf(text="""
            MDomain not-forbidden.org auto mail.not-forbidden.org

            <VirtualHost *:%s>
                ServerName not-forbidden.org
                ServerAlias test3.not-forbidden.org
                ServerAlias test4.not-forbidden.org
                SSLEngine on
            </VirtualHost>
            """ % TestEnv.HTTPS_PORT).install()
        assert TestEnv.apache_restart() == 0
        assert (0, 0) == TestEnv.httpd_error_log_count()

    # test case: MDomain does not match any vhost
    def test_300_012(self):
        HttpdConf(text="""
            MDomain example012.org www.example012.org
            <VirtualHost *:12346>
                ServerName not-forbidden.org
                ServerAlias test3.not-forbidden.org
            </VirtualHost>
            """).install()
        assert TestEnv.apache_restart() == 0
        assert (0, 1) == TestEnv.httpd_error_log_count()

    # test case: one md covers two vhosts
    def test_300_013(self):
        HttpdConf(text="""
            MDomain example2.org test-a.example2.org test-b.example2.org
            <VirtualHost *:12346>
                ServerName test-a.example2.org
            </VirtualHost>
            <VirtualHost *:12346>
                ServerName test-b.example2.org
            </VirtualHost>
            """).install()
        assert TestEnv.apache_restart() == 0
        assert (0, 0) == TestEnv.httpd_error_log_count()

    # test case: global server name as managed domain name
    def test_300_014(self):
        HttpdConf(text="""
            MDomain %s www.example2.org

            <VirtualHost *:12346>
                ServerName www.example2.org
            </VirtualHost>
            """ % TestEnv.HOSTNAME).install()
        assert TestEnv.apache_restart() == 0
        assert (0, 0) == TestEnv.httpd_error_log_count()

    # test case: valid pkey specification
    def test_300_015(self):
        HttpdConf(text="""
            MDPrivateKeys Default
            MDPrivateKeys RSA
            MDPrivateKeys RSA 2048
            MDPrivateKeys RSA 3072
            MDPrivateKeys RSA 4096
            """).install()
        assert TestEnv.apache_restart() == 0
        assert (0, 0) == TestEnv.httpd_error_log_count()

    # test case: invalid pkey specification
    @pytest.mark.parametrize("line,exp_err_msg", [
        ("MDPrivateKeys", "needs to specify the private key type"), 
        ("MDPrivateKeys Default RSA 1024", "'Default' allows no other parameter"),
        ("MDPrivateKeys RSA 1024", "must be 2048 or higher"),
        ("MDPrivateKeys RSA 1024", "must be 2048 or higher"),
        ("MDPrivateKeys rsa 2048 rsa 4096", "two keys of type 'RSA' are not possible"),
        ("MDPrivateKeys p-256 secp384r1 P-256", "two keys of type 'P-256' are not possible"),
        ])
    def test_300_016(self, line, exp_err_msg):
        HttpdConf(text=line).install()
        assert TestEnv.apache_restart() == 1
        assert exp_err_msg in TestEnv.apachectl_stderr

    # test case: invalid renew window directive
    @pytest.mark.parametrize("line,exp_err_msg", [
        ("MDRenewWindow dec-31", "has unrecognized format"), 
        ("MDRenewWindow 1y", "has unrecognized format"), 
        ("MDRenewWindow 10 d", "takes one argument"), 
        ("MDRenewWindow 102%", "a length of 100% or more is not allowed.")])
    def test_300_017(self, line, exp_err_msg):
        HttpdConf(text=line).install()
        assert TestEnv.apache_restart() == 1
        assert exp_err_msg in TestEnv.apachectl_stderr

    # test case: invalid uri for MDProxyPass
    @pytest.mark.parametrize("line,exp_err_msg", [
        ("MDHttpProxy", "takes one argument"), 
        ("MDHttpProxy localhost:8080", "scheme must be http or https"),
        ("MDHttpProxy https://127.0.0.1:-443", "invalid port"),
        ("MDHttpProxy HTTP localhost 8080", "takes one argument")])
    def test_300_018(self, line, exp_err_msg):
        HttpdConf(text=line).install()
        assert TestEnv.apache_restart() == 1, "Server accepted test config {}".format(line)
        assert exp_err_msg in TestEnv.apachectl_stderr

    # test case: invalid parameter for MDRequireHttps
    @pytest.mark.parametrize("line,exp_err_msg", [
        ("MDRequireHTTPS yes", "supported parameter values are 'temporary' and 'permanent'"),
        ("MDRequireHTTPS", "takes one argument")])
    def test_300_019(self, line, exp_err_msg):
        HttpdConf(text=line).install()
        assert TestEnv.apache_restart() == 1, "Server accepted test config {}".format(line)
        assert exp_err_msg in TestEnv.apachectl_stderr

    # test case: invalid parameter for MDMustStaple
    @pytest.mark.parametrize("line,exp_err_msg", [
        ("MDMustStaple", "takes one argument"), 
        ("MDMustStaple yes", "supported parameter values are 'on' and 'off'"),
        ("MDMustStaple true", "supported parameter values are 'on' and 'off'")])
    def test_300_020(self, line, exp_err_msg):
        HttpdConf(text=line).install()
        assert TestEnv.apache_restart() == 1, "Server accepted test config {}".format(line)
        assert exp_err_msg in TestEnv.apachectl_stderr

    # test case: alt-names incomplete detection, github isse #68
    def test_300_021(self):
        HttpdConf(text="""
            MDMembers manual
            MDomain secret.com
            <VirtualHost *:12344>
                ServerName not.secret.com
                ServerAlias secret.com
                SSLEngine on
            </VirtualHost>
            """).install()
        assert TestEnv.apache_fail() == 0
        assert (1, 0) == TestEnv.httpd_error_log_count()
        assert TestEnv.httpd_error_log_scan(
            re.compile(".*Virtual Host not.secret.com:0 matches Managed Domain 'secret.com', "
                       "but the name/alias not.secret.com itself is not managed. A requested "
                       "MD certificate will not match ServerName.*"))

    # test case: use MDRequireHttps in an <if> construct, but not in <Directory
    def test_300_022(self):
        HttpdConf(text="""
            MDomain secret.com
            <If "1 == 1">
              MDRequireHttps temporary
            </If>
            <VirtualHost *:12344>
                ServerName secret.com
                SSLEngine on
            </VirtualHost>
            """).install()
        assert TestEnv.apache_start() == 0
        HttpdConf(text="""
            MDomain secret.com
            <Directory /tmp>
              MDRequireHttps temporary
            </Directory>
            <VirtualHost *:12344>
                ServerName secret.com
                SSLEngine on
            </VirtualHost>
            """).install()
        assert TestEnv.apache_restart() == 1



# test mod_md acme terms-of-service handling

import os

from TestEnv import TestEnv
from TestHttpdConf import HttpdConf


def setup_module(module):
    print("setup_module: %s" % module.__name__)
    TestEnv.init()
    HttpdConf().install()
    assert TestEnv.apache_stop() == 0


def teardown_module(module):
    print("teardown_module:%s" % module.__name__)
    assert TestEnv.apache_stop() == 0
        

class TestStoreMigrate:

    # install old store, start a2md list, check files afterwards
    def test_0010_000(self):
        domain = "7007-1502285564.org"
        TestEnv.replace_store(os.path.join(TestEnv.TESTROOT, "data/store_migrate/1.0/sample1"))
        #
        # use 1.0 file name for private key
        fpkey_1_0 = os.path.join(TestEnv.STORE_DIR, 'domains', domain, 'pkey.pem')
        fpkey_1_1 = os.path.join(TestEnv.STORE_DIR, 'domains', domain, 'privkey.pem')
        cert_1_0 = os.path.join(TestEnv.STORE_DIR, 'domains', domain, 'cert.pem')
        cert_1_1 = os.path.join(TestEnv.STORE_DIR, 'domains', domain, 'pubcert.pem')
        chain_1_0 = os.path.join(TestEnv.STORE_DIR, 'domains', domain, 'chain.pem')
        #
        assert os.path.exists(fpkey_1_0)
        assert os.path.exists(cert_1_0)
        assert os.path.exists(chain_1_0)
        assert not os.path.exists(fpkey_1_1)
        assert not os.path.exists(cert_1_1)
        #
        md = TestEnv.a2md(["-vvv", "list", domain])['jout']['output'][0]
        assert domain == md["name"]
        #
        assert not os.path.exists(fpkey_1_0)
        assert os.path.exists(cert_1_0)
        assert os.path.exists(chain_1_0)
        assert os.path.exists(fpkey_1_1)
        assert os.path.exists(cert_1_1)



# test mod_md notify support

import os
from TestEnv import TestEnv
from TestHttpdConf import HttpdConf


class TestNotify:
    notify_cmd = None
    notify_log = None
    domain = None

    @classmethod
    def setup_class(cls):
        print("setup_class:%s" % cls.__name__)
        TestEnv.init()
        TestEnv.clear_store()
        TestEnv.check_acme()
        cls.domain = TestEnv.get_class_domain(cls)
        cls.notify_cmd = ("%s/notify.py" % TestEnv.TESTROOT)
        cls.notify_log = ("%s/notify.log" % TestEnv.GEN_DIR)

    @classmethod
    def teardown_class(cls):
        print("teardown_class:%s" % cls.__name__)
        assert TestEnv.apache_stop() == 0
    
    @classmethod
    def configure_httpd(cls, domain, add_lines=""):
        cls.domain = domain 
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_line(add_lines)
        conf.add_md([domain])
        conf.add_vhost(domain)
        conf.install()
        return domain
    
    def setup_method(self, method):
        TestNotify.domain = TestEnv.get_method_domain(method)
        if os.path.isfile(TestNotify.notify_log):
            os.remove(TestNotify.notify_log)

    # test: invalid notify cmd, check error
    def test_900_001(self):
        domain = TestNotify.domain
        command = "blablabla"
        args = ""
        TestNotify.configure_httpd(domain, """
            MDNotifyCmd %s %s
            """ % (command, args))
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain], restart=False)
        stat = TestEnv.get_md_status(domain)
        assert stat["renewal"]["last"]["problem"] == "urn:org:apache:httpd:log:AH10108:"

    # test: valid notify cmd that fails, check error
    def test_900_002(self):
        domain = TestNotify.domain
        command = "%s/notifail.py" % TestEnv.TESTROOT
        args = ""
        TestNotify.configure_httpd(domain, """
            MDNotifyCmd %s %s
            """ % (command, args))
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain], restart=False)
        stat = TestEnv.get_md_status(domain)
        assert stat["renewal"]["last"]["problem"] == "urn:org:apache:httpd:log:AH10108:"

    # test: valid notify that logs to file
    def test_900_010(self):
        domain = TestNotify.domain
        command = TestNotify.notify_cmd
        args = TestNotify.notify_log
        TestNotify.configure_httpd(domain, """
            MDNotifyCmd %s %s
            """ % (command, args))
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain], restart=False)
        stat = TestEnv.get_md_status(domain)
        assert stat["renewal"]["last"]["status"] == 0
        nlines = open(TestNotify.notify_log).readlines()
        assert 1 == len(nlines)
        assert ("['%s', '%s', '%s']" % (command, args, domain)) == nlines[0].strip()

    # test: signup with working notify cmd and see that it is called with the 
    #       configured extra arguments
    def test_900_011(self):
        domain = TestNotify.domain
        command = TestNotify.notify_cmd
        args = TestNotify.notify_log
        extra_arg = "test_900_011_extra"
        TestNotify.configure_httpd(domain, """
            MDNotifyCmd %s %s %s
            """ % (command, args, extra_arg))
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain], restart=False)
        stat = TestEnv.get_md_status(domain)
        assert stat["renewal"]["last"]["status"] == 0
        nlines = open(TestNotify.notify_log).readlines()
        assert ("['%s', '%s', '%s', '%s']" % (command, args, extra_arg, domain)) == nlines[0].strip()

    # test: signup with working notify cmd for 2 MD and expect it to be called twice
    def test_900_012(self):
        md1 = "a-" + TestNotify.domain
        domains1 = [md1, "www." + md1]
        md2 = "b-" + TestNotify.domain
        domains2 = [md2, "www." + md2]
        command = TestNotify.notify_cmd
        args = TestNotify.notify_log
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_notify_cmd("%s %s" % (command, args))
        conf.add_md(domains1)
        conf.add_md(domains2)
        conf.add_vhost(domains1)
        conf.add_vhost(domains2)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([md1, md2], restart=False)
        stat = TestEnv.get_md_status(md1)
        assert stat["renewal"]["last"]["status"] == 0
        stat = TestEnv.get_md_status(md2)
        assert stat["renewal"]["last"]["status"] == 0
        nlines = open(args).readlines()
        assert 2 == len(nlines)



#!/usr/bin/env python

import sys


def main(argv):
    if len(argv) > 2:
        f1 = open(argv[1], 'a+')
        f1.write('%s\n' % argv)
        f1.close()
        sys.stderr.write("done, all fine.\n")
        sys.exit(0)
    else:
        sys.stderr.write("%s without arguments" % (argv[0]))
        sys.exit(7)
    

if __name__ == "__main__":
    main(sys.argv)



# test mod_md status resources

import os
import re

from TestEnv import TestEnv
from TestHttpdConf import HttpdConf
from shutil import copyfile


def setup_module(module):
    print("setup_module    module:%s" % module.__name__)
    TestEnv.init()
    TestEnv.APACHE_CONF_SRC = "data/test_auto"
    TestEnv.check_acme()
    TestEnv.clear_store()
    HttpdConf().install()
    

def teardown_module(module):
    print("teardown_module module:%s" % module.__name__)
    assert TestEnv.apache_stop() == 0


class TestStatus:

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        TestEnv.clear_store()
        self.test_domain = TestEnv.get_method_domain(method)

    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    # simple MD, drive it, check status before activation
    def test_920_001(self):
        domain = self.test_domain
        domains = [domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_md(domains)
        conf.add_vhost(domain)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain], restart=False)
        # we started without a valid certificate, so we expect /.httpd/certificate-status
        # to not give information about one and - since we waited for the ACME signup
        # to complete - to give information in 'renewal' about the new cert.
        status = TestEnv.get_certificate_status(domain)
        assert 'sha256-fingerprint' not in status
        assert 'valid' not in status
        assert 'renewal' in status
        assert 'valid' in status['renewal']['cert']
        assert 'sha256-fingerprint' in status['renewal']['cert']['rsa']
        # restart and activate
        # once activated, the staging must be gone and attributes exist for the active cert
        assert TestEnv.apache_restart() == 0
        status = TestEnv.get_certificate_status(domain)
        assert 'renewal' not in status
        assert 'sha256-fingerprint' in status['rsa']
        assert 'valid' in status['rsa']
        assert 'from' in status['rsa']['valid']

    # simple MD, drive it, manipulate staged credentials and check status
    def test_920_002(self):
        domain = self.test_domain
        domains = [domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_md(domains)
        conf.add_vhost(domain)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain], restart=False)
        # copy a real certificate from LE over to staging
        staged_cert = os.path.join(TestEnv.STORE_DIR, 'staging', domain, 'pubcert.pem') 
        real_cert = os.path.join('data', 'test_920', '002.pubcert')
        assert copyfile(real_cert, staged_cert)
        status = TestEnv.get_certificate_status(domain)
        # status shows the copied cert's properties as staged
        assert 'renewal' in status
        assert 'Thu, 29 Aug 2019 16:06:35 GMT' == status['renewal']['cert']['rsa']['valid']['until']
        assert 'Fri, 31 May 2019 16:06:35 GMT' == status['renewal']['cert']['rsa']['valid']['from']
        assert '03039C464D454EDE79FCD2CAE859F668F269' == status['renewal']['cert']['rsa']['serial']
        assert 'sha256-fingerprint' in status['renewal']['cert']['rsa']

    # test if switching status off has effect
    def test_920_003(self):
        domain = self.test_domain
        domains = [domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_md(domains)
        conf.add_line("MDCertificateStatus off")
        conf.add_vhost(domain)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain], restart=False)
        status = TestEnv.get_certificate_status(domain)
        assert not status

    # get the complete md-status JSON, check that it
    def test_920_004(self):
        domain = self.test_domain
        domains = [domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_md(domains)
        conf.add_line("MDCertificateStatus off")
        conf.add_vhost(domain)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain])
        status = TestEnv.get_md_status("")
        assert "version" in status
        assert "managed-domains" in status
        assert 1 == len(status["managed-domains"])

    # get the status of a domain on base server
    def test_920_010(self):
        domain = self.test_domain
        domains = [domain]
        conf = HttpdConf(std_vhosts=False, text="""
LogLevel md:trace2
LogLevel ssl:debug
                
MDBaseServer on
MDPortMap http:- https:%s

Listen %s
ServerAdmin admin@not-forbidden.org
ServerName %s
SSLEngine on
Protocols h2 http/1.1 acme-tls/1

<Location "/server-status">
    SetHandler server-status
</Location>
<Location "/md-status">
    SetHandler md-status
</Location>
            """ % (TestEnv.HTTPS_PORT, TestEnv.HTTPS_PORT, domain))
        conf.add_md(domains)
        conf.install()
        TestEnv.HTTPD_CHECK_URL = TestEnv.HTTPD_URL_SSL
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain], restart=False)
        status = TestEnv.get_md_status("")
        assert "version" in status
        assert "managed-domains" in status
        assert 1 == len(status["managed-domains"])
        # get the html page
        status = TestEnv.get_server_status()
        assert re.search(r'<h3>Managed Certificates</h3>', status, re.MULTILINE)
        # get the ascii summary
        status = TestEnv.get_server_status(query="?auto")
        m = re.search(r'Managed Certificates: total=(\d+), ok=(\d+) renew=(\d+) errored=(\d+) ready=(\d+)',
                      status, re.MULTILINE)
        assert 1 == int(m.group(1))
        assert 0 == int(m.group(2))
        assert 1 == int(m.group(3))
        assert 0 == int(m.group(4))
        assert 1 == int(m.group(5))

    def test_920_011(self):
        # MD with static cert files in base server, see issue #161
        domain = self.test_domain
        domains = [domain, 'www.%s' % domain]
        testpath = os.path.join(TestEnv.GEN_DIR, 'test_920_011')
        # cert that is only 10 more days valid
        TestEnv.create_self_signed_cert(domains, {"notBefore": -70, "notAfter": 20},
                                        serial=920011, path=testpath)
        cert_file = os.path.join(testpath, 'pubcert.pem')
        pkey_file = os.path.join(testpath, 'privkey.pem')
        assert os.path.exists(cert_file)
        assert os.path.exists(pkey_file)
        conf = HttpdConf(std_vhosts=False, text=f"""
LogLevel md:trace2
LogLevel ssl:debug
                
MDPortMap http:- https:{TestEnv.HTTPS_PORT}

Listen {TestEnv.HTTPS_PORT}
ServerAdmin admin@not-forbidden.org
ServerName {domain}
SSLEngine on
Protocols h2 http/1.1 acme-tls/1

MDBaseServer on

<Location "/server-status">
    SetHandler server-status
</Location>
<Location "/md-status">
    SetHandler md-status
</Location>
            """)
        conf.start_md(domains)
        conf.add_line(f"MDCertificateFile {cert_file}")
        conf.add_line(f"MDCertificateKeyFile {pkey_file}")
        conf.end_md()
        conf.install()
        TestEnv.HTTPD_CHECK_URL = TestEnv.HTTPD_URL_SSL
        assert TestEnv.apache_restart() == 0
        status = TestEnv.get_md_status(domain)
        assert status
        assert 'renewal' not in status
        print(status)
        assert status['state'] == TestEnv.MD_S_COMPLETE
        assert status['renew-mode'] == 1  # manual

    # MD with 2 certificates
    def test_920_020(self):
        domain = self.test_domain
        domains = [domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_line("MDStapling on")
        conf.add_line("MDPrivateKeys secp256r1 RSA")
        conf.add_md(domains)
        conf.add_vhost(domain)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain], restart=False)
        # In the stats JSON, we excpect 2 certificates under 'renewal'
        stat = TestEnv.get_md_status(domain)
        assert 'renewal' in stat
        assert 'cert' in stat['renewal']
        assert 'rsa' in stat['renewal']['cert']
        assert 'secp256r1' in stat['renewal']['cert']
        # In /.httpd/certificate-status 'renewal' we excpect 2 certificates
        status = TestEnv.get_certificate_status(domain)
        assert 'renewal' in status
        assert 'cert' in status['renewal']
        assert 'secp256r1' in status['renewal']['cert']
        assert 'rsa' in status['renewal']['cert']
        # restart and activate
        # once activated, certs are listed in status
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.get_md_status(domain)
        assert 'cert' in stat
        assert 'valid' in stat['cert']
        for ktype in ['rsa', 'secp256r1']:
            assert ktype in stat['cert']
            if not TestEnv.ACME_LACKS_OCSP:
                assert 'ocsp' in stat['cert'][ktype]



# test mod_md acme terms-of-service handling

import os

from shutil import copyfile
from TestEnv import TestEnv


def setup_module(module):
    print("setup_module: %s" % module.__name__)
    TestEnv.init()


def teardown_module(module):
    print("teardown_module: %s" % module.__name__)


class TestRegAdd:

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        TestEnv.clear_store()

    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    # test case: list empty store
    def test_120_000(self):
        assert TestEnv.a2md(["list"])['jout'] == TestEnv.EMPTY_JOUT

    # test case: list two managed domains
    def test_120_001(self):
        domains = [ 
            ["test120-001.com", "test120-001a.com", "test120-001b.com"],
            ["greenbytes2.de", "www.greenbytes2.de", "mail.greenbytes2.de"]
        ]
        for dns in domains:
            assert TestEnv.a2md(["add"] + dns)['rv'] == 0
        #
        # list all store content
        jout = TestEnv.a2md(["list"])['jout']
        assert len(jout['output']) == len(domains)
        domains.reverse()
        for i in range(0, len(jout['output'])):
            TestEnv.check_json_contains(jout['output'][i], {
                "name": domains[i][0],
                "domains": domains[i],
                "contacts": [],
                "ca": {
                    "url": TestEnv.ACME_URL,
                    "proto": "ACME"
                },
                "state": TestEnv.MD_S_INCOMPLETE
            })
        # list md by name
        for dns in ["test120-001.com", "greenbytes2.de"]:
            md = TestEnv.a2md(["list", dns])['jout']['output'][0]
            assert md['name'] == dns

    # test case: validate md state in store
    def test_120_002(self):
        # check: md without pkey/cert -> INCOMPLETE
        domain = "not-forbidden.org"
        assert TestEnv.a2md(["add", domain])['rv'] == 0
        assert TestEnv.a2md(["update", domain, "contacts", "admin@" + domain])['rv'] == 0
        assert TestEnv.a2md(["update", domain, "agreement", TestEnv.ACME_TOS])['rv'] == 0
        assert TestEnv.a2md(["list", domain])['jout']['output'][0]['state'] == TestEnv.MD_S_INCOMPLETE
        # check: valid pkey/cert -> COMPLETE
        copyfile(self._path_conf_ssl("valid_pkey.pem"), TestEnv.store_domain_file(domain, 'privkey.pem'))
        copyfile(self._path_conf_ssl("valid_cert.pem"), TestEnv.store_domain_file(domain, 'pubcert.pem'))
        assert TestEnv.a2md(["list", domain])['jout']['output'][0]['state'] == TestEnv.MD_S_COMPLETE
        # check: expired cert -> EXPIRED
        copyfile(self._path_conf_ssl("expired_pkey.pem"), TestEnv.store_domain_file(domain, 'privkey.pem'))
        copyfile(self._path_conf_ssl("expired_cert.pem"), TestEnv.store_domain_file(domain, 'pubcert.pem'))
        out = TestEnv.a2md(["list", domain])['jout']['output'][0]
        assert out['state'] == TestEnv.MD_S_INCOMPLETE
        assert out['renew'] is True

    # test case: broken cert file
    def test_120_003(self):
        domain = "not-forbidden.org"
        assert TestEnv.a2md(["add", domain])['rv'] == 0
        assert TestEnv.a2md(["update", domain, "contacts", "admin@" + domain])['rv'] == 0
        assert TestEnv.a2md(["update", domain, "agreement", TestEnv.ACME_TOS])['rv'] == 0
        # check: valid pkey/cert -> COMPLETE
        copyfile(self._path_conf_ssl("valid_pkey.pem"), TestEnv.store_domain_file(domain, 'privkey.pem'))
        copyfile(self._path_conf_ssl("valid_cert.pem"), TestEnv.store_domain_file(domain, 'pubcert.pem'))
        assert TestEnv.a2md(["list", domain])['jout']['output'][0]['state'] == TestEnv.MD_S_COMPLETE
        # check: replace cert by broken file -> ERROR
        copyfile(self._path_conf_ssl("valid_cert.req"), TestEnv.store_domain_file(domain, 'pubcert.pem'))
        assert TestEnv.a2md(["list", domain])['jout']['output'][0]['state'] == TestEnv.MD_S_ERROR

    # REMOVED: we no longer verify private keys at startup and leave that to the
    #          user of the key, e.g. mod_ssl. It is the ultimate arbiter of this.
    # def test_120_004(self):
    # test case: broken private key file

    def _path_conf_ssl(self, name):
        return os.path.join(TestEnv.APACHE_SSL_DIR, name) 



# test auto runs against ACMEv2

import os
import pytest

from TestEnv import TestEnv
from TestHttpdConf import HttpdConf
from TestCertUtil import CertUtil


def setup_module(module):
    print("setup_module    module:%s" % module.__name__)
    TestEnv.init()
    TestEnv.APACHE_CONF_SRC = "data/test_auto"
    TestEnv.check_acme()
    TestEnv.clear_store()
    HttpdConf().install()
    assert TestEnv.apache_start() == 0


def teardown_module(module):
    print("teardown_module module:%s" % module.__name__)
    assert TestEnv.apache_stop() == 0


class TestAutov2:

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        TestEnv.httpd_error_log_clear()
        TestEnv.clear_store()
        self.test_domain = TestEnv.get_method_domain(method)

    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    # create a MD not used in any virtual host, auto drive should NOT pick it up
    def test_702_001(self):
        domain = self.test_domain
        # generate config with one MD
        domains = [domain, "www." + domain]
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_drive_mode("auto")
        conf.add_md(domains)
        conf.install()
        #
        # restart, check that MD is synched to store
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)
        stat = TestEnv.get_md_status(domain)
        assert stat["watched"] == 0
        #
        # add vhost for MD, restart should drive it
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain])
        TestEnv.check_md_complete(domain)
        stat = TestEnv.get_md_status(domain)
        assert stat["watched"] == 1
        cert = TestEnv.get_cert(domain)
        assert domain in cert.get_san_list()
        #
        # challenges should have been removed
        # file system needs to have correct permissions
        TestEnv.check_dir_empty(TestEnv.store_challenges())
        TestEnv.check_file_permissions(domain)

    # test case: same as test_7001, but with two parallel managed domains
    def test_702_002(self):
        domain = self.test_domain
        domain_a = "a-" + domain
        domain_b = "b-" + domain
        #        
        # generate config with two MDs
        domains_a = [domain_a, "www." + domain_a]
        domains_b = [domain_b, "www." + domain_b]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_drive_mode("auto")
        conf.add_md(domains_a)
        conf.add_md(domains_b)
        conf.add_vhost(domains_a)
        conf.add_vhost(domains_b)
        conf.install()
        #
        # restart, check that md is in store
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains_a)
        TestEnv.check_md(domains_b)
        #
        # await drive completion, do not restart
        assert TestEnv.await_completion([domain_a, domain_b], restart=False)
        # staged certificates are now visible on the status resources
        status = TestEnv.get_md_status(domain_a)
        assert 'renewal' in status
        assert 'cert' in status['renewal']
        assert 'rsa' in status['renewal']['cert']
        assert 'sha256-fingerprint' in status['renewal']['cert']['rsa']
        # restart and activate
        assert TestEnv.apache_restart() == 0
        # check: SSL is running OK
        cert_a = TestEnv.get_cert(domain_a)
        assert domains_a == cert_a.get_san_list()
        cert_b = TestEnv.get_cert(domain_b)
        assert domains_b == cert_b.get_san_list()

    # test case: one MD, that covers two vhosts
    def test_702_003(self):
        domain = self.test_domain
        name_a = "test-a." + domain
        name_b = "test-b." + domain
        domains = [domain, name_a, name_b]
        #
        # generate 1 MD and 2 vhosts
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_md(domains)
        conf.add_vhost(name_a, doc_root="htdocs/a")
        conf.add_vhost(name_b, doc_root="htdocs/b")
        conf.install()
        #
        # create docRoot folder
        self._write_res_file(os.path.join(TestEnv.APACHE_HTDOCS_DIR, "a"), "name.txt", name_a)
        self._write_res_file(os.path.join(TestEnv.APACHE_HTDOCS_DIR, "b"), "name.txt", name_b)
        #
        # restart (-> drive), check that MD was synched and completes
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)
        assert TestEnv.await_completion([domain])
        TestEnv.check_md_complete(domain)
        #
        # check: SSL is running OK
        cert_a = TestEnv.get_cert(name_a)
        assert name_a in cert_a.get_san_list()
        cert_b = TestEnv.get_cert(name_b)
        assert name_b in cert_b.get_san_list()
        assert cert_a.same_serial_as(cert_b)
        #
        assert TestEnv.get_content(name_a, "/name.txt") == name_a
        assert TestEnv.get_content(name_b, "/name.txt") == name_b

    # test case: drive with using single challenge type explicitly
    @pytest.mark.parametrize("challenge_type", [
        "tls-alpn-01", "http-01"
    ])
    def test_702_004(self, challenge_type):
        domain = self.test_domain
        domains = [domain, "www." + domain]
        #
        # generate 1 MD and 1 vhost
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_line("Protocols http/1.1 acme-tls/1")
        conf.add_drive_mode("auto")
        conf.add_ca_challenges([challenge_type])
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        #
        # restart (-> drive), check that MD was synched and completes
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)
        assert TestEnv.await_completion([domain])
        TestEnv.check_md_complete(domain)
        #        
        # check SSL running OK
        cert = TestEnv.get_cert(domain)
        assert domain in cert.get_san_list()

    # test case: drive_mode manual, check that server starts, but requests to domain are 503'd
    def test_702_005(self):
        domain = self.test_domain
        name_a = "test-a." + domain
        domains = [domain, name_a]
        #
        # generate 1 MD and 1 vhost
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_drive_mode("manual")
        conf.add_md(domains)
        conf.add_vhost(name_a, doc_root="htdocs/a")
        conf.install()
        #
        # create docRoot folder
        self._write_res_file(os.path.join(TestEnv.APACHE_HTDOCS_DIR, "a"), "name.txt", name_a)
        #
        # restart, check that md is in store
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)
        #        
        # check: that request to domains give 503 Service Unavailable
        cert1 = TestEnv.get_cert(name_a)
        assert name_a in cert1.get_san_list()
        assert TestEnv.getStatus(name_a, "/name.txt") == 503
        #
        # check temporary cert from server
        cert2 = CertUtil(TestEnv.path_fallback_cert(domain))
        assert cert1.same_serial_as(cert2), \
            "Unexpected temporary certificate on vhost %s. Expected cn: %s , "\
            "but found cn: %s" % (name_a, cert2.get_cn(), cert1.get_cn())

    # test case: drive MD with only invalid challenges, domains should stay 503'd
    def test_702_006(self):
        domain = self.test_domain
        name_a = "test-a." + domain
        domains = [domain, name_a]
        #
        # generate 1 MD, 1 vhost
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_ca_challenges(["invalid-01", "invalid-02"])
        conf.add_md(domains)
        conf.add_vhost(name_a, doc_root="htdocs/a")
        conf.install()
        #
        # create docRoot folder
        self._write_res_file(os.path.join(TestEnv.APACHE_HTDOCS_DIR, "a"), "name.txt", name_a)
        #
        # restart, check that md is in store
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)
        # await drive completion
        md = TestEnv.await_error(domain)
        assert md
        assert md['renewal']['errors'] > 0
        assert md['renewal']['last']['problem'] == 'challenge-mismatch'
        assert 'account' not in md['ca']
        #
        # check: that request to domains give 503 Service Unavailable
        cert = TestEnv.get_cert(name_a)
        assert name_a in cert.get_san_list()
        assert TestEnv.getStatus(name_a, "/name.txt") == 503

    # Specify a non-working http proxy
    def test_702_008(self):
        domain = self.test_domain
        domains = [domain]
        #
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_drive_mode("always")
        conf.add_http_proxy("http://localhost:1")
        conf.add_md(domains)
        conf.install()
        #
        # - restart (-> drive)
        assert TestEnv.apache_restart() == 0
        # await drive completion
        md = TestEnv.await_error(domain)
        assert md
        assert md['renewal']['errors'] > 0
        assert md['renewal']['last']['status-description'] == 'Connection refused'
        assert 'account' not in md['ca']

    # Specify a valid http proxy
    def test_702_008a(self):
        domain = self.test_domain
        domains = [domain]
        #
        conf = HttpdConf(proxy=True)
        conf.add_admin("admin@" + domain)
        conf.add_drive_mode("always")
        conf.add_http_proxy("http://localhost:%s" % TestEnv.HTTP_PROXY_PORT)
        conf.add_md(domains)
        conf.install()
        #
        # - restart (-> drive), check that md is in store
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain])
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md_complete(domain)

    # Force cert renewal due to critical remaining valid duration
    # Assert that new cert activation is delayed
    def test_702_009(self):
        domain = self.test_domain
        domains = [domain]
        #
        # prepare md
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_drive_mode("auto")
        conf.add_renew_window("10d")
        conf.add_md(domains)
        conf.add_vhost(domain)
        conf.install()
        #
        # restart (-> drive), check that md+cert is in store, TLS is up
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain])
        TestEnv.check_md_complete(domain)
        cert1 = CertUtil(TestEnv.store_domain_file(domain, 'pubcert.pem'))
        # compare with what md reports as status
        stat = TestEnv.get_certificate_status(domain)
        assert cert1.same_serial_as(stat['rsa']['serial'])
        #
        # create self-signed cert, with critical remaining valid duration -> drive again
        TestEnv.create_self_signed_cert([domain], {"notBefore": -120, "notAfter": 2}, serial=7029)
        cert3 = CertUtil(TestEnv.store_domain_file(domain, 'pubcert.pem'))
        assert cert3.same_serial_as('1B75')
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.get_certificate_status(domain)
        assert cert3.same_serial_as(stat['rsa']['serial'])
        #
        # cert should renew and be different afterwards
        assert TestEnv.await_completion([domain], must_renew=True)
        stat = TestEnv.get_certificate_status(domain)
        assert not cert3.same_serial_as(stat['rsa']['serial'])
        
    # test case: drive with an unsupported challenge due to port availability 
    def test_702_010(self):
        domain = self.test_domain
        domains = [domain, "www." + domain]
        #
        # generate 1 MD and 1 vhost, map port 80 to where the server does not listen
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_drive_mode("auto")
        conf._add_line("MDPortMap 80:99")        
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        md = TestEnv.await_error(domain)
        assert md["renewal"]["errors"] > 0
        #
        # now the same with a 80 mapped to a supported port 
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_drive_mode("auto")
        conf.add_ca_challenges(["http-01"])
        conf._add_line("MDPortMap 80:%s" % TestEnv.HTTP_PORT)
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)
        assert TestEnv.await_completion([domain])

    def test_702_011(self):
        domain = self.test_domain
        domains = [domain, "www." + domain]
        #
        # generate 1 MD and 1 vhost, map port 443 to where the server does not listen
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_line("Protocols http/1.1 acme-tls/1")
        conf.add_drive_mode("auto")
        conf._add_line("MDPortMap https:99 http:99")        
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        md = TestEnv.await_error(domain)
        assert md["renewal"]["errors"] > 0
        #
        # now the same with a 443 mapped to a supported port 
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_line("Protocols http/1.1 acme-tls/1")
        conf.add_drive_mode("auto")
        conf.add_ca_challenges(["tls-alpn-01"])
        conf._add_line("MDPortMap https:%s" % TestEnv.HTTPS_PORT)
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)
        assert TestEnv.await_completion([domain])

    # test case: one MD with several dns names. sign up. remove the *first* name
    # in the MD. restart. should find and keep the existing MD.
    # See: https://github.com/icing/mod_md/issues/68
    def test_702_030(self):
        domain = self.test_domain
        name_x = "test-x." + domain
        name_a = "test-a." + domain
        name_b = "test-b." + domain
        domains = [name_x, name_a, name_b]
        #
        # generate 1 MD and 2 vhosts
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_md(domains)
        conf.add_vhost(name_a)
        conf.add_vhost(name_b)
        conf.install()
        #
        # restart (-> drive), check that MD was synched and completes
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)
        assert TestEnv.await_completion([name_x])
        TestEnv.check_md_complete(name_x)
        #
        # check: SSL is running OK
        cert_a = TestEnv.get_cert(name_a)
        assert name_a in cert_a.get_san_list()
        cert_b = TestEnv.get_cert(name_b)
        assert name_b in cert_b.get_san_list()
        assert cert_a.same_serial_as(cert_b)
        #        
        # change MD by removing 1st name
        new_list = [name_a, name_b]
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_md(new_list)
        conf.add_vhost(name_a)
        conf.add_vhost(name_b)
        conf.install()
        # restart, check that host still works and kept the cert
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(new_list)
        status = TestEnv.get_certificate_status(name_a)
        assert cert_a.same_serial_as(status['rsa']['serial'])

    # test case: Same as 7030, but remove *and* add another at the same time.
    # restart. should find and keep the existing MD and renew for additional name.
    # See: https://github.com/icing/mod_md/issues/68
    def test_702_031(self):
        domain = self.test_domain
        name_x = "test-x." + domain
        name_a = "test-a." + domain
        name_b = "test-b." + domain
        name_c = "test-c." + domain
        domains = [name_x, name_a, name_b]
        #
        # generate 1 MD and 2 vhosts
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_md(domains)
        conf.add_vhost(name_a)
        conf.add_vhost(name_b)
        conf.install()
        #
        # restart (-> drive), check that MD was synched and completes
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)
        assert TestEnv.await_completion([name_x])
        TestEnv.check_md_complete(name_x)
        #
        # check: SSL is running OK
        cert_a = TestEnv.get_cert(name_a)
        assert name_a in cert_a.get_san_list()
        cert_b = TestEnv.get_cert(name_b)
        assert name_b in cert_b.get_san_list()
        assert cert_a.same_serial_as(cert_b)
        #        
        # change MD by removing 1st name and adding another
        new_list = [name_a, name_b, name_c]
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_md(new_list)
        conf.add_vhost(name_a)
        conf.add_vhost(name_b)
        conf.install()
        # restart, check that host still works and have new cert
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(new_list)
        assert TestEnv.await_completion([name_a])
        #
        cert_a2 = TestEnv.get_cert(name_a)
        assert name_a in cert_a2.get_san_list()
        assert not cert_a.same_serial_as(cert_a2)

    # test case: create two MDs, move them into one
    # see: <https://bz.apache.org/bugzilla/show_bug.cgi?id=62572>
    def test_702_032(self):
        domain = self.test_domain
        name1 = "server1." + domain
        name2 = "server2.b" + domain  # need a separate TLD to avoid rate limites
        #
        # generate 2 MDs and 2 vhosts
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf._add_line("MDMembers auto")
        conf.add_md([name1])
        conf.add_md([name2])
        conf.add_vhost(name1)
        conf.add_vhost(name2)
        conf.install()
        #
        # restart (-> drive), check that MD was synched and completes
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md([name1])
        TestEnv.check_md([name2])
        assert TestEnv.await_completion([name1, name2])
        TestEnv.check_md_complete(name2)
        #
        # check: SSL is running OK
        cert1 = TestEnv.get_cert(name1)
        assert name1 in cert1.get_san_list()
        cert2 = TestEnv.get_cert(name2)
        assert name2 in cert2.get_san_list()
        #        
        # remove second md and vhost, add name2 to vhost1
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf._add_line("MDMembers auto")
        conf.add_md([name1])
        conf.add_vhost([name1, name2])
        conf.install()
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md([name1, name2])
        assert TestEnv.await_completion([name1])
        #
        cert1b = TestEnv.get_cert(name1)
        assert name1 in cert1b.get_san_list()
        assert name2 in cert1b.get_san_list()
        assert not cert1.same_serial_as(cert1b)

    # test case: test "tls-alpn-01" challenge handling
    def test_702_040(self):
        domain = self.test_domain
        domains = [domain, "www." + domain]
        #
        # generate 1 MD and 1 vhost
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_line("LogLevel core:debug")
        conf.add_line("LogLevel ssl:debug")
        conf.add_line("Protocols http/1.1 acme-tls/1")
        conf.add_drive_mode("auto")
        conf.add_ca_challenges(["tls-alpn-01"])
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        #
        # restart (-> drive), check that MD was synched and completes
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)
        # check that acme-tls/1 is available for all domains
        stat = TestEnv.get_md_status(domain)
        assert stat["proto"]["acme-tls/1"] == domains
        assert TestEnv.await_completion([domain])
        TestEnv.check_md_complete(domain)
        #        
        # check SSL running OK
        cert = TestEnv.get_cert(domain)
        assert domain in cert.get_san_list()

    # test case: test "tls-alpn-01" without enabling 'acme-tls/1' challenge protocol
    def test_702_041(self):
        domain = self.test_domain
        domains = [domain, "www." + domain]
        #
        # generate 1 MD and 1 vhost
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_line("LogLevel core:debug")
        conf.add_line("LogLevel ssl:debug")
        conf.add_drive_mode("auto")
        conf.add_ca_challenges(["tls-alpn-01"])
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        #
        # restart (-> drive), check that MD job shows errors 
        # and that missing proto is detected
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)
        # check that acme-tls/1 is available for none of the domains
        stat = TestEnv.get_md_status(domain)
        assert stat["proto"]["acme-tls/1"] == []

    # test case: 2.4.40 mod_ssl stumbles over a SSLCertificateChainFile when installing
    # a fallback certificate
    def test_702_042(self):
        domain = self.test_domain
        dns_list = [domain]
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_line("LogLevel core:debug")
        conf.add_line("LogLevel ssl:debug")
        conf.add_line("SSLCertificateChainFile %s" % (self._path_conf_ssl("valid_cert.pem")))
        conf.add_drive_mode("auto")
        conf.add_md(dns_list)
        conf.add_vhost(TestEnv.HTTPS_PORT, dns_list)
        conf.install()
        assert TestEnv.apache_restart() == 0
        
    # Make a setup using the base server. It will use http-01 challenge.
    def test_702_050(self):
        domain = self.test_domain
        conf = HttpdConf()
        conf.add_line("""
            MDBaseServer on
            ServerAdmin admin@%s
            ServerName %s
            """ % (domain, domain))
        conf.add_md([domain])
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain])

    # Make a setup using the base server without http:, will fail.
    def test_702_051(self):
        domain = self.test_domain
        conf = HttpdConf()
        conf.add_line("""
            MDBaseServer on
            MDPortMap http:-
            ServerAdmin admin@%s
            ServerName %s
            """ % (domain, domain))
        conf.add_md([domain])
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_error(domain)

    # Make a setup using the base server without http:, but with acme-tls/1, should work.
    def test_702_052(self):
        domain = self.test_domain
        conf = HttpdConf()
        conf.add_line("""
            MDBaseServer on
            MDPortMap http:-
            Protocols h2 http/1.1 acme-tls/1
            ServerAdmin admin@%s
            ServerName %s
            SSLEngine on
            """ % (domain, domain))
        conf.add_md([domain])
        conf.install()
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.get_md_status(domain)
        assert stat["proto"]["acme-tls/1"] == [domain]
        assert TestEnv.await_completion([domain])

    # Test a domain name longer than 64 chars, but components < 64, see #227
    # Background: DNS has an official limit of 253 ASCII chars and components must be
    # of length [1, 63].
    # However the CN in a certificate is restricted too, see
    # <https://github.com/letsencrypt/boulder/issues/2093>.
    @pytest.mark.skipif(TestEnv.ACME_SERVER == 'pebble', reason="pebble differs here from boulder")
    @pytest.mark.parametrize("challenge_type", [
        "tls-alpn-01", "http-01"
    ])
    def test_702_060(self, challenge_type):
        domain = self.test_domain
        # use only too long names, this is expected to fail:
        # see <https://github.com/jetstack/cert-manager/issues/1462>
        long_domain = ("x" * (65 - len(domain))) + domain
        domains = [long_domain, "www." + long_domain]
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_line("Protocols http/1.1 acme-tls/1")
        conf.add_drive_mode("auto")
        conf.add_ca_challenges([challenge_type])
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)
        assert TestEnv.await_error(long_domain)
        # add a short domain to the SAN list, the CA should now use that one
        # and issue a cert.
        long_domain = ("y" * (65 - len(domain))) + domain
        domains = [long_domain, "www." + long_domain, "xxx." + domain]
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_line("Protocols http/1.1 acme-tls/1")
        conf.add_drive_mode("auto")
        conf.add_ca_challenges([challenge_type])
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([long_domain])
        TestEnv.check_md_complete(long_domain)
        #
        # check SSL running OK
        cert = TestEnv.get_cert(long_domain)
        assert long_domain in cert.get_san_list()

    # --------- _utils_ ---------

    def _write_res_file(self, doc_root, name, content):
        if not os.path.exists(doc_root):
            os.makedirs(doc_root)
        open(os.path.join(doc_root, name), "w").write(content)

    def _path_conf_ssl(self, name):
        return os.path.join(TestEnv.APACHE_SSL_DIR, name) 



# tests with elliptic curve keys and certificates

import pytest

from TestEnv import TestEnv
from TestHttpdConf import HttpdConf


def setup_module(module):
    print("setup_module    module:%s" % module.__name__)
    TestEnv.init()
    TestEnv.APACHE_CONF_SRC = "data/test_auto"
    TestEnv.check_acme()
    TestEnv.clear_store()
    HttpdConf().install()
    assert TestEnv.apache_start() == 0


def teardown_module(module):
    print("teardown_module module:%s" % module.__name__)
    assert TestEnv.apache_stop() == 0


class TestAutov2:

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        TestEnv.httpd_error_log_clear()
        TestEnv.clear_store()
        self.test_domain = TestEnv.get_method_domain(method)

    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    def set_get_pkeys(self, domain, pkeys, conf=None):
        domains = [domain]
        if conf is None:
            conf = HttpdConf()
            conf.add_admin("admin@" + domain)
            conf.add_line("MDPrivateKeys {0}".format(" ".join([p['spec'] for p in pkeys])))
            conf.add_md(domains)
            conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain])

    def check_pkeys(self, domain, pkeys):
        # check that files for all types have been created
        for p in [p for p in pkeys if len(p['spec'])]:
            TestEnv.check_md_complete(domain, p['spec'])
        # check that openssl client sees the cert with given keylength for cipher
        TestEnv.verify_cert_key_lenghts(domain, pkeys)
    
    def set_get_check_pkeys(self, domain, pkeys, conf=None):
        self.set_get_pkeys(domain, pkeys, conf=conf)
        self.check_pkeys(domain, pkeys)
        
    # one EC key, no RSA
    def test_810_001(self):
        domain = self.test_domain
        self.set_get_check_pkeys(domain, [
            {'spec': "secp256r1", 'ciphers': "ECDSA", 'keylen': 256},
            {'spec': "", 'ciphers': "RSA", 'keylen': 0},
        ])

    # set EC key type override on MD and get certificate
    def test_810_002(self):
        domain = self.test_domain
        # generate config with one MD
        domains = [domain]
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_line("MDPrivateKeys secp256r1")
        conf.start_md(domains)
        conf.add_line("    MDPrivateKeys secp384r1")
        conf.end_md()
        conf.add_vhost(domains)
        self.set_get_check_pkeys(domain, [ 
            {'spec': "secp384r1", 'ciphers': "ECDSA", 'keylen': 384},
            {'spec': "", 'ciphers': "RSA", 'keylen': 0},
        ])

    # set two key spec, ec before rsa
    def test_810_003a(self):
        domain = self.test_domain
        self.set_get_check_pkeys(domain, [ 
            {'spec': "P-256", 'ciphers': "ECDSA", 'keylen': 256},
            {'spec': "RSA 3072", 'ciphers': "RSA", 'keylen': 3072},
        ])

    # set two key spec, rsa before ec
    def test_810_003b(self):
        domain = self.test_domain
        self.set_get_check_pkeys(domain, [ 
            {'spec': "RSA 3072", 'ciphers': "RSA", 'keylen': 3072},
            {'spec': "secp384r1", 'ciphers': "ECDSA", 'keylen': 384},
        ])

    # use a curve unsupported by LE
    def test_810_004(self):
        domain = self.test_domain
        # generate config with one MD
        domains = [domain]
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_line("MDPrivateKeys secp192r1")
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        md = TestEnv.await_error(domain)
        assert md
        assert md['renewal']['errors'] > 0
        assert md['renewal']['last']['problem'] == 'urn:ietf:params:acme:error:malformed'

    # set three key specs
    def test_810_005(self):
        domain = self.test_domain
        self.set_get_check_pkeys(domain, [ 
            {'spec': "secp256r1", 'ciphers': "ECDSA", 'keylen': 384},  # we will see the cert from 3
            {'spec': "RSA 4096", 'ciphers': "RSA", 'keylen': 4096},
            {'spec': "P-384", 'ciphers': "ECDSA", 'keylen': 384},
        ])

    # set three key specs
    def test_810_006(self):
        domain = self.test_domain
        self.set_get_check_pkeys(domain, [
            {'spec': "rsa2048", 'ciphers': "RSA", 'keylen': 2048},
            {'spec': "secp256r1", 'ciphers': "ECDSA", 'keylen': 256},
        ])

    # disabled completely, since LE does not support that key type
    # X25529 key type which has some special quirks
    #@pytest.mark.skip(reason="this is not working yet.")
    #def test_810_010(self):
    #    domain = self.test_domain
    #    self.set_get_check_pkeys(domain, [
    #        {'spec': "x25519", 'ciphers': "ECDSA", 'keylen': 384},
    #    ])



# test mod_md ACMEv2 registrations

import re
import json
import pytest

from TestEnv import TestEnv


def setup_module(module):
    print("setup_module: %s" % module.__name__)
    TestEnv.init()


def teardown_module(module):
    print("teardown_module:%s" % module.__name__)


class TestAcmeAcc:

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        TestEnv.check_acme()
        TestEnv.clear_store()
 
    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    # test case: register a new account, vary length to check base64 encoding
    @pytest.mark.parametrize("contact", [
        "x@not-forbidden.org", "xx@not-forbidden.org", "xxx@not-forbidden.org"
    ])
    def test_202_000(self, contact):
        run = TestEnv.a2md(["-t", "accepted", "acme", "newreg", contact], raw=True)
        assert run['rv'] == 0
        m = re.match("registered: (.*)$", run["stdout"])
        assert m, "did not match: {0}".format(run["stdout"])
        acct = m.group(1)
        print("newreg: %s" % m.group(1))
        self._check_account(acct, ["mailto:" + contact])

    # test case: register a new account without accepting ToS, must fail
    def test_202_000b(self):
        run = TestEnv.a2md(["acme", "newreg", "x@not-forbidden.org"], raw=True)
        assert run['rv'] == 1
        m = re.match(".*must agree to terms of service.*", run["stderr"])
        if m is None:
            # the pebble variant
            m = re.match(".*account did not agree to the terms of service.*", run["stderr"])
        assert m, "did not match: {0}".format(run["stderr"])

    # test case: respect 'mailto:' prefix in contact url
    def test_202_001(self):
        contact = "mailto:xx@not-forbidden.org"
        run = TestEnv.a2md(["-t", "accepted", "acme", "newreg", contact], raw=True)
        assert run['rv'] == 0
        m = re.match("registered: (.*)$", run["stdout"])
        assert m
        acct = m.group(1)
        self._check_account(acct, [contact])

    # test case: fail on invalid contact url
    @pytest.mark.parametrize("invalid_contact", [
        "mehlto:xxx@not-forbidden.org", "no.at.char", "with blank@test.com",
        "missing.host@", "@missing.localpart.de",
        "double..dot@test.com", "double@at@test.com"
    ])
    def test_202_002(self, invalid_contact):
        assert TestEnv.a2md(["acme", "newreg", invalid_contact])['rv'] == 1

    # test case: use contact list
    def test_202_003(self):
        contact = ["xx@not-forbidden.org", "aa@not-forbidden.org"]
        run = TestEnv.a2md(["-t", "accepted", "acme", "newreg"] + contact, raw=True)
        assert run['rv'] == 0
        m = re.match("registered: (.*)$", run["stdout"])
        assert m
        acct = m.group(1)
        self._check_account(acct, ["mailto:" + contact[0], "mailto:" + contact[1]])

    # test case: validate new account
    def test_202_100(self):
        acct = self._prepare_account(["tmp@not-forbidden.org"])
        assert TestEnv.a2md(["acme", "validate", acct])['rv'] == 0

    # test case: fail on non-existing account
    def test_202_101(self):
        assert TestEnv.a2md(["acme", "validate", "ACME-localhost-1000"])['rv'] == 1

    # test case: report fail on request signing problem
    def test_202_102(self):
        acct = self._prepare_account(["tmp@not-forbidden.org"])
        with open(TestEnv.path_account(acct)) as f:
            acctj = json.load(f)
        acctj['url'] = acctj['url'] + "0"
        open(TestEnv.path_account(acct), "w").write(json.dumps(acctj))
        assert TestEnv.a2md(["acme", "validate", acct])['rv'] == 1

    # test case: register and try delete an account, will fail without persistence
    def test_202_200(self):
        acct = self._prepare_account(["tmp@not-forbidden.org"])
        assert TestEnv.a2md(["delreg", acct])['rv'] == 1

    # test case: register and try delete an account with persistence
    def test_202_201(self):
        acct = self._prepare_account(["tmp@not-forbidden.org"])
        assert TestEnv.a2md(["acme", "delreg", acct])['rv'] == 0
        # check that store is clean
        run = TestEnv.run(["find", TestEnv.STORE_DIR])
        assert re.match(TestEnv.STORE_DIR, run['stdout'])

    # test case: delete a persisted account without specifying url
    def test_202_202(self):
        acct = self._prepare_account(["tmp@not-forbidden.org"])
        assert TestEnv.run([TestEnv.A2MD, "-d", TestEnv.STORE_DIR, "acme", "delreg", acct])['rv'] == 0

    # test case: delete, then validate an account
    def test_202_203(self):
        acct = self._prepare_account(["test014@not-forbidden.org"])
        assert TestEnv.a2md(["acme", "delreg", acct])['rv'] == 0
        # validate on deleted account fails
        assert TestEnv.a2md(["acme", "validate", acct])['rv'] == 1

    def _check_account(self, acct, contact):
        with open(TestEnv.path_account(acct)) as f:
            acctj = json.load(f)
        assert acctj['registration']['contact'] == contact

    def _prepare_account(self, contact):
        run = TestEnv.a2md(["-t", "accepted", "acme", "newreg"] + contact, raw=True)
        assert run['rv'] == 0
        return re.match("registered: (.*)$", run['stdout']).group(1)



# test mod_md message support

import json
import os
import re
import time
import pytest

from TestEnv import TestEnv
from TestHttpdConf import HttpdConf


def setup_module(module):
    print("setup_module    module:%s" % module.__name__)
    TestEnv.init()
    TestEnv.APACHE_CONF_SRC = "data/test_auto"
    TestEnv.check_acme()
    TestEnv.clear_store()
    HttpdConf().install()
    assert TestEnv.apache_start() == 0
    

def teardown_module(module):
    print("teardown_module module:%s" % module.__name__)
    assert TestEnv.apache_stop() == 0


class TestMessage:

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        TestEnv.clear_store()
        self.test_domain = TestEnv.get_method_domain(method)
        self.mcmd = ("%s/message.py" % TestEnv.TESTROOT)
        self.mcmdfail = ("%s/notifail.py" % TestEnv.TESTROOT)
        self.mlog = ("%s/message.log" % TestEnv.GEN_DIR)
        if os.path.isfile(self.mlog):
            os.remove(self.mlog)

    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    # test: signup with configured message cmd that is invalid
    def test_901_001(self):
        domain = self.test_domain
        domains = [domain, "www." + domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_message_cmd("blablabla")
        conf.add_drive_mode("auto")
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_file(TestEnv.store_staged_file(domain, 'job.json'))
        stat = TestEnv.get_md_status(domain)
        # this command should have failed and logged an error
        assert stat["renewal"]["last"]["problem"] == "urn:org:apache:httpd:log:AH10109:"

    # test: signup with configured message cmd that is valid but returns != 0
    def test_901_002(self):
        self.mcmd = ("%s/notifail.py" % TestEnv.TESTROOT)
        domain = self.test_domain
        domains = [domain, "www." + domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_message_cmd("%s %s" % (self.mcmd, self.mlog))
        conf.add_drive_mode("auto")
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain], restart=False)
        stat = TestEnv.get_md_status(domain)
        # this command should have failed and logged an error
        assert stat["renewal"]["last"]["problem"] == "urn:org:apache:httpd:log:AH10109:"

    # test: signup with working message cmd and see that it logs the right things
    def test_901_003(self):
        domain = self.test_domain
        domains = [domain, "www." + domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_message_cmd("%s %s" % (self.mcmd, self.mlog))
        conf.add_drive_mode("auto")
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain], restart=False)
        stat = TestEnv.get_md_status(domain)
        # this command did not fail and logged itself the correct information
        assert stat["renewal"]["last"]["status"] == 0
        assert stat["renewal"]["log"]["entries"]
        assert stat["renewal"]["log"]["entries"][0]["type"] == "message-renewed"
        # shut down server to make sure that md has completed 
        assert TestEnv.apache_stop() == 0
        nlines = open(self.mlog).readlines()
        assert 3 == len(nlines)
        nlines = [s.strip() for s in nlines]
        assert "['{cmd}', '{logfile}', 'challenge-setup:http-01:{dns}', '{mdomain}']".format(
            cmd=self.mcmd, logfile=self.mlog, mdomain=domain, dns=domains[0]) in nlines
        assert "['{cmd}', '{logfile}', 'challenge-setup:http-01:{dns}', '{mdomain}']".format(
            cmd=self.mcmd, logfile=self.mlog, mdomain=domain, dns=domains[1]) in nlines
        assert nlines[2].strip() == "['{cmd}', '{logfile}', 'renewed', '{mdomain}']".format(
            cmd=self.mcmd, logfile=self.mlog, mdomain=domain)

    # test issue #145: 
    # - a server renews a valid certificate and is not restarted when recommended
    # - the job did not clear its next_run and was run over and over again
    # - the job logged the re-verifications again and again. which was saved.
    # - this eventually flushed out the "message-renew" log entry
    # - which caused the renew message handling to trigger again and again
    # the fix does:
    # - reset the next run
    # - no longer adds the re-validations to the log
    # - messages only once
    @pytest.mark.skipif(TestEnv.ACME_SERVER == 'pebble', reason="ACME server certs valid too long")
    def test_901_004(self):
        domain = self.test_domain
        domains = [domain, "www." + domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain])
        # force renew
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_message_cmd("%s %s" % (self.mcmd, self.mlog))
        conf.add_line("MDRenewWindow 120d")
        conf.add_line("MDActivationDelay -7d")
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain], restart=False)
        TestEnv.get_md_status(domain)
        assert TestEnv.await_file(self.mlog)
        nlines = open(self.mlog).readlines()
        assert 1 == len(nlines)
        assert ("['%s', '%s', 'renewed', '%s']" % (self.mcmd, self.mlog, domain)) == nlines[0].strip()
    
    def test_901_010(self):
        # MD with static cert files, lifetime in renewal window, no message about renewal
        domain = self.test_domain
        domains = [domain, 'www.%s' % domain]
        testpath = os.path.join(TestEnv.GEN_DIR, 'test_901_010')
        # cert that is only 10 more days valid
        TestEnv.create_self_signed_cert(domains, {"notBefore": -70, "notAfter": 20},
                                        serial=901010, path=testpath)
        cert_file = os.path.join(testpath, 'pubcert.pem')
        pkey_file = os.path.join(testpath, 'privkey.pem')
        assert os.path.exists(cert_file)
        assert os.path.exists(pkey_file)
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_message_cmd("%s %s" % (self.mcmd, self.mlog))
        conf.start_md(domains)
        conf.add_line("MDCertificateFile %s" % cert_file)
        conf.add_line("MDCertificateKeyFile %s" % pkey_file)
        conf.end_md()
        conf.add_vhost(domain)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert not os.path.isfile(self.mlog)
        
    def test_901_011(self):
        # MD with static cert files, lifetime in warn window, check message
        domain = self.test_domain
        domains = [domain, 'www.%s' % domain]
        testpath = os.path.join(TestEnv.GEN_DIR, 'test_901_011')
        # cert that is only 10 more days valid
        TestEnv.create_self_signed_cert(domains, {"notBefore": -85, "notAfter": 5},
                                        serial=901011, path=testpath)
        cert_file = os.path.join(testpath, 'pubcert.pem')
        pkey_file = os.path.join(testpath, 'privkey.pem')
        assert os.path.exists(cert_file)
        assert os.path.exists(pkey_file)
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_message_cmd("%s %s" % (self.mcmd, self.mlog))
        conf.start_md(domains)
        conf.add_line("MDCertificateFile %s" % cert_file)
        conf.add_line("MDCertificateKeyFile %s" % pkey_file)
        conf.end_md()
        conf.add_vhost(domain)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_file(self.mlog)
        nlines = open(self.mlog).readlines()
        assert 1 == len(nlines)
        assert ("['%s', '%s', 'expiring', '%s']" % (self.mcmd, self.mlog, domain)) == nlines[0].strip()
        # check that we do not get it resend right away again
        assert TestEnv.apache_restart() == 0
        time.sleep(1)
        nlines = open(self.mlog).readlines()
        assert 1 == len(nlines)
        assert ("['%s', '%s', 'expiring', '%s']" % (self.mcmd, self.mlog, domain)) == nlines[0].strip()

    # MD, check messages from stapling
    @pytest.mark.skipif(TestEnv.ACME_LACKS_OCSP, reason="no OCSP responder")
    def test_901_020(self):
        domain = self.test_domain
        domains = [domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_message_cmd("%s %s" % (self.mcmd, self.mlog))
        conf.add_drive_mode("auto")
        conf.add_md(domains)
        conf.add_line("MDStapling on")
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain])
        TestEnv.await_ocsp_status(domain)
        assert TestEnv.await_file(self.mlog)
        time.sleep(1)
        nlines = open(self.mlog).readlines()
        assert 4 == len(nlines)
        assert nlines[0].strip() == ("['%s', '%s', 'challenge-setup:http-01:%s', '%s']"
                                     % (self.mcmd, self.mlog, domain, domain))
        assert nlines[1].strip() == ("['%s', '%s', 'renewed', '%s']" % (self.mcmd, self.mlog, domain))
        assert nlines[2].strip() == ("['%s', '%s', 'installed', '%s']" % (self.mcmd, self.mlog, domain))
        assert nlines[3].strip() ==  ("['%s', '%s', 'ocsp-renewed', '%s']" % (self.mcmd, self.mlog, domain))

    # test: while testing gh issue #146, it was noted that a failed renew notification never
    # resets the MD activity.
    @pytest.mark.skipif(TestEnv.ACME_SERVER == 'pebble', reason="ACME server certs valid too long")
    def test_901_030(self):
        domain = self.test_domain
        domains = [domain, "www." + domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain])
        # set the warn window that triggers right away and a failing message command
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_message_cmd("%s %s" % (self.mcmdfail, self.mlog))
        conf.add_md(domains)
        conf.add_line("""
            MDWarnWindow 100d
            """)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        TestEnv.get_md_status(domain)
        # this command should have failed and logged an error
        # shut down server to make sure that md has completed
        assert TestEnv.await_file(TestEnv.store_staged_file(domain, 'job.json'))
        while True:
            with open(TestEnv.store_staged_file(domain, 'job.json')) as f:
                job = json.load(f)
                if job["errors"] > 0:
                    assert job["errors"] > 0,  "unexpected job result: {0}".format(job)
                    assert job["last"]["problem"] == "urn:org:apache:httpd:log:AH10109:"
                    break
            time.sleep(0.1)

        # reconfigure to a working notification command and restart
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_message_cmd("%s %s" % (self.mcmd, self.mlog))
        conf.add_md(domains)
        conf.add_line("""
            MDWarnWindow 100d
            """)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_file(self.mlog)
        # we see the notification logged by the command
        nlines = open(self.mlog).readlines()
        assert 1 == len(nlines)
        assert ("['%s', '%s', 'expiring', '%s']" % (self.mcmd, self.mlog, domain)) == nlines[0].strip()
        # the error needs to be gone
        assert TestEnv.await_file(TestEnv.store_staged_file(domain, 'job.json'))
        with open(TestEnv.store_staged_file(domain, 'job.json')) as f:
            job = json.load(f)
            assert job["errors"] == 0



# test mod_md basic configurations

import os

from TestEnv import TestEnv
from TestHttpdConf import HttpdConf


def setup_module(module):
    print("setup_module    module:%s" % module.__name__)
    TestEnv.APACHE_CONF_SRC = "data/test_roundtrip"
    

def teardown_module(module):
    print("teardown_module module:%s" % module.__name__)
    assert TestEnv.apache_stop() == 0


class TestRoundtripv2:

    @classmethod
    def setup_class(cls):
        TestEnv.clear_store()
        HttpdConf().install()

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        TestEnv.check_acme()
        self.test_domain = TestEnv.get_method_domain(method)

    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    # --------- add to store ---------

    def test_602_000(self):
        # test case: generate config with md -> restart -> drive -> generate config
        # with vhost and ssl -> restart -> check HTTPS access
        domain = self.test_domain
        domains = [domain, "www." + domain]

        # - generate config with one md
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_drive_mode("manual")
        conf.add_md(domains)
        conf.install()
        # - restart, check that md is in store
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)
        # - drive
        assert TestEnv.a2md(["-v", "drive", domain])['rv'] == 0
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md_complete(domain)
        # - append vhost to config
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        # check: SSL is running OK
        cert = TestEnv.get_cert(domain)
        assert domain in cert.get_san_list()

        # check file system permissions:
        TestEnv.check_file_permissions(domain)

    def test_602_001(self):
        # test case: same as test_600_000, but with two parallel managed domains
        domain_a = "a-" + self.test_domain
        domain_b = "b-" + self.test_domain
        # - generate config with one md
        domains_a = [domain_a, "www." + domain_a]
        domains_b = [domain_b, "www." + domain_b]

        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_drive_mode("manual")
        conf.add_md(domains_a)
        conf.add_md(domains_b)
        conf.install()

        # - restart, check that md is in store
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains_a)
        TestEnv.check_md(domains_b)

        # - drive
        assert TestEnv.a2md(["drive", domain_a])['rv'] == 0
        assert TestEnv.a2md(["drive", domain_b])['rv'] == 0
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md_complete(domain_a)
        TestEnv.check_md_complete(domain_b)

        # - append vhost to config
        conf.add_vhost(domains_a)
        conf.add_vhost(domains_b)
        conf.install()

        # check: SSL is running OK
        assert TestEnv.apache_restart() == 0
        cert_a = TestEnv.get_cert(domain_a)
        assert domains_a == cert_a.get_san_list()
        cert_b = TestEnv.get_cert(domain_b)
        assert domains_b == cert_b.get_san_list()

    def test_602_002(self):
        # test case: one md, that covers two vhosts
        domain = self.test_domain
        name_a = "a." + domain
        name_b = "b." + domain
        domains = [domain, name_a, name_b]

        # - generate config with one md
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_drive_mode("manual")
        conf.add_md(domains)
        conf.install()
        
        # - restart, check that md is in store
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md(domains)

        # - drive
        assert TestEnv.a2md(["drive", domain])['rv'] == 0
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md_complete(domain)

        # - append vhost to config
        conf.add_vhost(name_a, doc_root="htdocs/a")
        conf.add_vhost(name_b, doc_root="htdocs/b")
        conf.install()
        
        # - create docRoot folder
        self._write_res_file(os.path.join(TestEnv.APACHE_HTDOCS_DIR, "a"), "name.txt", name_a)
        self._write_res_file(os.path.join(TestEnv.APACHE_HTDOCS_DIR, "b"), "name.txt", name_b)

        # check: SSL is running OK
        assert TestEnv.apache_restart() == 0
        cert_a = TestEnv.get_cert(name_a)
        assert name_a in cert_a.get_san_list()
        cert_b = TestEnv.get_cert(name_b)
        assert name_b in cert_b.get_san_list()
        assert cert_a.same_serial_as(cert_b)
        assert TestEnv.get_content(name_a, "/name.txt") == name_a
        assert TestEnv.get_content(name_b, "/name.txt") == name_b

    # --------- _utils_ ---------

    def _write_res_file(self, doc_root, name, content):
        if not os.path.exists(doc_root):
            os.makedirs(doc_root)
        open(os.path.join(doc_root, name), "w").write(content)



# test mod_md must-staple support
import pytest

from TestEnv import TestEnv
from TestHttpdConf import HttpdConf
from TestCertUtil import CertUtil


class TestMustStaple:
    domain = None

    @classmethod
    def setup_class(cls):
        print("setup_class:%s" % cls.__name__)
        TestEnv.init()
        TestEnv.clear_store()
        TestEnv.check_acme()
        cls.domain = TestEnv.get_class_domain(cls)
        cls.configure_httpd(cls.domain)
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([cls.domain])

    @classmethod
    def teardown_class(cls):
        print("teardown_class:%s" % cls.__name__)
        assert TestEnv.apache_stop() == 0

    @classmethod
    def configure_httpd(cls, domain, add_lines=""):
        cls.domain = domain 
        conf = HttpdConf()
        conf.add_admin("admin@" + domain)
        conf.add_line(add_lines)
        conf.add_md([domain])
        conf.add_vhost(domain)
        conf.install()
        return domain
    
    # MD with default, e.g. not staple
    def test_800_001(self):
        domain = TestMustStaple.domain
        TestEnv.check_md_complete(domain)
        cert1 = CertUtil(TestEnv.store_domain_file(domain, 'pubcert.pem'))
        assert not cert1.get_must_staple()

    # MD that should explicitly not staple
    def test_800_002(self):
        domain = TestMustStaple.domain
        TestMustStaple.configure_httpd(domain, "MDMustStaple off")
        assert TestEnv.apache_restart() == 0
        TestEnv.check_md_complete(domain)
        cert1 = CertUtil(TestEnv.store_domain_file(domain, 'pubcert.pem'))
        assert not cert1.get_must_staple()
        stat = TestEnv.get_ocsp_status(domain)
        assert stat['ocsp'] == "no response sent" 

    # MD that must staple and toggle off again
    @pytest.mark.skipif(TestEnv.ACME_LACKS_OCSP, reason="no OCSP responder")
    def test_800_003(self):
        domain = TestMustStaple.domain
        TestMustStaple.configure_httpd(domain, "MDMustStaple on")
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain])
        TestEnv.check_md_complete(domain)
        cert1 = CertUtil(TestEnv.store_domain_file(domain, 'pubcert.pem'))
        assert cert1.get_must_staple()
        domain = TestMustStaple.configure_httpd(domain, "MDMustStaple off")
        assert TestEnv.apache_restart() == 0
        assert TestEnv.await_completion([domain])
        TestEnv.check_md_complete(domain)
        cert1 = CertUtil(TestEnv.store_domain_file(domain, 'pubcert.pem'))
        assert not cert1.get_must_staple()

    # MD that must staple
    @pytest.mark.skipif(TestEnv.ACME_LACKS_OCSP, reason="no OCSP responder")
    def test_800_004(self):
        domain = TestMustStaple.domain
        # mod_ssl stapling is off, expect no stapling
        stat = TestEnv.get_ocsp_status(domain)
        assert stat['ocsp'] == "no response sent" 
        # turn mod_ssl stapling on, expect an answer
        domain = TestMustStaple.configure_httpd(domain, """
            LogLevel ssl:trace2
            SSLUseStapling On
            SSLStaplingCache \"shmcb:logs/ssl_stapling(32768)\"
            """)
        assert TestEnv.apache_restart() == 0
        stat = TestEnv.get_ocsp_status(domain)
        assert stat['ocsp'] == "successful (0x0)" 
        assert stat['verify'] == "0 (ok)"



# test ACME error responses and their processing

from TestEnv import TestEnv
from TestHttpdConf import HttpdConf


def setup_module(module):
    print("setup_module    module:%s" % module.__name__)
    TestEnv.init()
    TestEnv.APACHE_CONF_SRC = "data/test_auto"
    TestEnv.check_acme()
    TestEnv.clear_store()
    HttpdConf().install()
    assert TestEnv.apache_start() == 0
    

def teardown_module(module):
    print("teardown_module module:%s" % module.__name__)
    assert TestEnv.apache_stop() == 0


class TestAcmeErrors:

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        self.test_domain = TestEnv.get_method_domain(method)

    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    # -----------------------------------------------------------------------------------------------
    # test case: MD with 2 names, one invalid
    #
    def test_740_000(self):
        domain = self.test_domain
        domains = [domain, "invalid!." + domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        md = TestEnv.await_error(domain)
        assert md
        assert md['renewal']['errors'] > 0
        if TestEnv.ACME_SERVER == 'pebble':
            assert md['renewal']['last']['problem'] == 'urn:ietf:params:acme:error:malformed'
            assert md['renewal']['last']['detail'] == "Order included DNS identifier with a value containing an illegal character: '!'"
        else:
            assert md['renewal']['last']['problem'] == 'urn:ietf:params:acme:error:rejectedIdentifier'
            assert md['renewal']['last']['detail'] == (
                    "Error creating new order :: Cannot issue for "
                    "\"%s\": Domain name contains an invalid character" % domains[1])

    # test case: MD with 3 names, 2 invalid
    #
    def test_740_001(self):
        domain = self.test_domain
        domains = [domain, "invalid1!." + domain, "invalid2!." + domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_md(domains)
        conf.add_vhost(domains)
        conf.install()
        assert TestEnv.apache_restart() == 0
        md = TestEnv.await_error(domain)
        assert md
        assert md['renewal']['errors'] > 0
        if TestEnv.ACME_SERVER == 'pebble':
            assert md['renewal']['last']['problem'] == 'urn:ietf:params:acme:error:malformed'
            assert md['renewal']['last']['detail'].startswith(
                "Order included DNS identifier with a value containing an illegal character")
        else:
            assert md['renewal']['last']['problem'] == 'urn:ietf:params:acme:error:rejectedIdentifier'
            assert md['renewal']['last']['detail'].startswith(
                "Error creating new order :: Cannot issue for")
            assert md['renewal']['last']['subproblems']
            assert len(md['renewal']['last']['subproblems']) == 2



# test mod_md acme terms-of-service handling

import pytest

from TestEnv import TestEnv


def setup_module(module):
    print("setup_module: %s" % module.__name__)
    TestEnv.init()


def teardown_module(module):
    print("teardown_module: %s" % module.__name__)


class TestRegAdd:

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        TestEnv.clear_store()
 
    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    # test case: add a single dns managed domain
    def test_100_000(self):
        dns = "greenbytes.de"
        jout1 = TestEnv.a2md(["add", dns])['jout']
        TestEnv.check_json_contains(jout1['output'][0], {
            "name": dns,
            "domains": [dns],
            "contacts": [],
            "ca": {
                "url": TestEnv.ACME_URL,
                "proto": "ACME"
            },
            "state": TestEnv.MD_S_INCOMPLETE
        })
        assert TestEnv.a2md(["list"])['jout'] == jout1

    # test case: add > 1 dns managed domain
    def test_100_001(self):
        dns = ["greenbytes2.de", "www.greenbytes2.de", "mail.greenbytes2.de"]
        jout1 = TestEnv.a2md(["add"] + dns)['jout']
        TestEnv.check_json_contains(jout1['output'][0], {
            "name": dns[0],
            "domains": dns,
            "contacts": [],
            "ca": {
                "url": TestEnv.ACME_URL,
                "proto": "ACME"
            },
            "state": TestEnv.MD_S_INCOMPLETE
        })
        assert TestEnv.a2md(["list"])['jout'] == jout1

    # test case: add second managed domain
    def test_100_002(self):
        dns1 = ["test100-002.com", "test100-002a.com", "test100-002b.com"]
        TestEnv.a2md(["add"] + dns1)
        # add second managed domain
        dns2 = ["greenbytes2.de", "www.greenbytes2.de", "mail.greenbytes2.de"]
        jout = TestEnv.a2md(["add"] + dns2)['jout']
        # assert: output covers only changed md
        assert len(jout['output']) == 1
        TestEnv.check_json_contains(jout['output'][0], {
            "name": dns2[0],
            "domains": dns2,
            "contacts": [],
            "ca": {
                "url": TestEnv.ACME_URL,
                "proto": "ACME"
            },
            "state": TestEnv.MD_S_INCOMPLETE
        })
        assert len(TestEnv.a2md(["list"])['jout']['output']) == 2

    # test case: add existing domain 
    def test_100_003(self):
        dns = "greenbytes.de"
        assert TestEnv.a2md(["add", dns])['rv'] == 0
        assert TestEnv.a2md(["add", dns])['rv'] == 1

    # test case: add without CA URL
    def test_100_004(self):
        dns = "greenbytes.de"
        jout1 = TestEnv.run([TestEnv.A2MD, "-d", TestEnv.STORE_DIR, "-j", "add", dns])['jout']
        assert len(jout1['output']) == 1
        TestEnv.check_json_contains(jout1['output'][0], {
            "name": dns,
            "domains": [dns],
            "contacts": [],
            "ca": {
                "proto": "ACME"
            },
            "state": TestEnv.MD_S_INCOMPLETE
        })
        assert TestEnv.a2md(["list"])['jout'] == jout1

    # test case: add with invalid DNS
    @pytest.mark.parametrize("invalid_dns", [
        "tld", "white sp.ace", "invalid.*.wildcard.com", "k\xc3ller.idn.com"
    ])
    def test_100_005(self, invalid_dns):
        assert TestEnv.a2md(["add", invalid_dns])["rv"] == 1
        assert TestEnv.a2md(["add", "test-100.de", invalid_dns])["rv"] == 1

    # test case: add with invalid ACME URL
    @pytest.mark.parametrize("invalid_url", [
        "no.schema/path", "http://white space/path", "http://bad.port:-1/path"])
    def test_100_006(self, invalid_url):
        args = [TestEnv.A2MD, "-a", invalid_url, "-d", TestEnv.STORE_DIR, "-j"]
        dns = "greenbytes.de"
        args.extend(["add", dns])
        assert TestEnv.run(args)["rv"] == 1

    # test case: add overlapping dns names
    def test_100_007(self):
        assert TestEnv.a2md(["add", "test-100.com", "test-101.com"])['rv'] == 0
        # 1: alternate DNS exists as primary name
        assert TestEnv.a2md(["add", "greenbytes2.de", "test-100.com"])['rv'] == 1
        # 2: alternate DNS exists as alternate DNS
        assert TestEnv.a2md(["add", "greenbytes2.de", "test-101.com"])['rv'] == 1
        # 3: primary name exists as alternate DNS
        assert TestEnv.a2md(["add", "test-101.com"])['rv'] == 1

    # test case: add subdomains as separate managed domain
    def test_100_008(self):
        assert TestEnv.a2md(["add", "test-100.com"])['rv'] == 0
        assert TestEnv.a2md(["add", "sub.test-100.com"])['rv'] == 0

    # test case: add duplicate domain
    def test_100_009(self):
        dns1 = "test-100.com"
        dns2 = "test-101.com"
        jout = TestEnv.a2md(["add", dns1, dns2, dns1, dns2])['jout']
        # DNS is only listed once
        assert len(jout['output']) == 1
        md = jout['output'][0]
        assert md['domains'] == [dns1, dns2]

    # test case: add pnuycode name
    def test_100_010(self):
        assert TestEnv.a2md(["add", "xn--kller-jua.punycode.de"])['rv'] == 0

    # test case: don't sort alternate names
    def test_100_011(self):
        dns = ["test-100.com", "test-xxx.com", "test-aaa.com"]
        jout = TestEnv.a2md(["add"] + dns)['jout']
        # DNS is only listed as specified
        assert len(jout['output']) == 1
        md = jout['output'][0]
        assert md['domains'] == dns

    # test case: add DNS wildcard
    @pytest.mark.parametrize("wild_dns", [
        "*.wildcard.com"
    ])
    def test_100_012(self, wild_dns):
        assert TestEnv.a2md(["add", wild_dns])['rv'] == 0



# test MDs with static certificates

import os

from TestEnv import TestEnv
from TestHttpdConf import HttpdConf


def setup_module(module):
    print("setup_module    module:%s" % module.__name__)
    TestEnv.init()
    TestEnv.APACHE_CONF_SRC = "data/test_auto"
    TestEnv.check_acme()
    TestEnv.clear_store()
    HttpdConf().install()
    

def teardown_module(module):
    print("teardown_module module:%s" % module.__name__)
    assert TestEnv.apache_stop() == 0


class TestStatus:

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        TestEnv.clear_store()
        self.test_domain = TestEnv.get_method_domain(method)

    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    def test_730_001(self):
        # MD with static cert files, will not be driven
        domain = self.test_domain
        domains = [domain, 'www.%s' % domain]
        testpath = os.path.join(TestEnv.GEN_DIR, 'test_920_001')
        # cert that is only 10 more days valid
        TestEnv.create_self_signed_cert(domains, {"notBefore": -80, "notAfter": 10},
                                        serial=730001, path=testpath)
        cert_file = os.path.join(testpath, 'pubcert.pem')
        pkey_file = os.path.join(testpath, 'privkey.pem')
        assert os.path.exists(cert_file)
        assert os.path.exists(pkey_file)
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.start_md(domains)
        conf.add_line("MDCertificateFile %s" % cert_file)
        conf.add_line("MDCertificateKeyFile %s" % pkey_file)
        conf.end_md()
        conf.add_vhost(domain)
        conf.install()
        assert TestEnv.apache_restart() == 0
        
        # check if the domain uses it, it appears in our stats and renewal is off
        cert = TestEnv.get_cert(domain)
        assert cert.same_serial_as(730001)
        stat = TestEnv.get_md_status(domain)
        assert stat
        assert 'cert' in stat
        assert stat['renew'] is True
        assert 'renewal' not in stat

    def test_730_002(self):
        # MD with static cert files, force driving
        domain = self.test_domain
        domains = [domain, 'www.%s' % domain]
        testpath = os.path.join(TestEnv.GEN_DIR, 'test_920_001')
        # cert that is only 10 more days valid
        TestEnv.create_self_signed_cert(domains, {"notBefore": -80, "notAfter": 10},
                                        serial=730001, path=testpath)
        cert_file = os.path.join(testpath, 'pubcert.pem')
        pkey_file = os.path.join(testpath, 'privkey.pem')
        assert os.path.exists(cert_file)
        assert os.path.exists(pkey_file)
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.start_md(domains)
        conf.add_line("MDCertificateFile %s" % cert_file)
        conf.add_line("MDCertificateKeyFile %s" % pkey_file)
        conf.add_line("MDRenewMode always")
        conf.end_md()
        conf.add_vhost(domain)
        conf.install()
        assert TestEnv.apache_restart() == 0
        
        # check if the domain uses it, it appears in our stats and renewal is off
        cert = TestEnv.get_cert(domain)
        assert cert.same_serial_as(730001)
        stat = TestEnv.get_md_status(domain)
        assert stat
        assert 'cert' in stat
        assert stat['renew'] is True
        assert TestEnv.await_renewal(domains)

    def test_730_003(self):
        # just configuring one file will not work
        domain = self.test_domain
        domains = [domain, 'www.%s' % domain]
        testpath = os.path.join(TestEnv.GEN_DIR, 'test_920_001')
        # cert that is only 10 more days valid
        TestEnv.create_self_signed_cert(domains, {"notBefore": -80, "notAfter": 10},
                                        serial=730001, path=testpath)
        cert_file = os.path.join(testpath, 'pubcert.pem')
        pkey_file = os.path.join(testpath, 'privkey.pem')
        assert os.path.exists(cert_file)
        assert os.path.exists(pkey_file)
        
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.start_md(domains)
        conf.add_line("MDCertificateFile %s" % cert_file)
        conf.end_md()
        conf.add_vhost(domain)
        conf.install()
        assert TestEnv.apache_fail() == 0
        
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.start_md(domains)
        conf.add_line("MDCertificateKeyFile %s" % pkey_file)
        conf.end_md()
        conf.add_vhost(domain)
        conf.install()
        assert TestEnv.apache_fail() == 0



# test mod_md cleanups and sanitation

import os
from TestEnv import TestEnv
from TestHttpdConf import HttpdConf


def setup_module(module):
    print("setup_module    module:%s" % module.__name__)
    TestEnv.init()
    TestEnv.APACHE_CONF_SRC = "data/test_auto"
    TestEnv.check_acme()
    TestEnv.clear_store()
    HttpdConf().install()


def teardown_module(module):
    print("teardown_module module:%s" % module.__name__)
    assert TestEnv.apache_stop() == 0


class TestCleanups:

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        TestEnv.clear_store()
        self.test_domain = TestEnv.get_method_domain(method)

    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    def test_910_01(self):
        # generate a simple MD
        domain = self.test_domain
        domains = [domain]
        conf = HttpdConf()
        conf.add_admin("admin@not-forbidden.org")
        conf.add_drive_mode("manual")
        conf.add_md(domains)
        conf.add_vhost(domain)
        conf.install()

        # create valid/invalid challenges subdirs
        challenges_dir = TestEnv.store_challenges()
        dirs_before = ["aaa", "bbb", domain, "zzz"]
        for name in dirs_before:
            os.makedirs(os.path.join(challenges_dir, name))

        assert TestEnv.apache_restart() == 0
        # the one we use is still there
        assert os.path.isdir(os.path.join(challenges_dir, domain))
        # and the others are gone
        missing_after = ["aaa", "bbb", "zzz"]
        for name in missing_after:
            assert not os.path.exists(os.path.join(challenges_dir, name))



# test mod_md acme terms-of-service handling

import pytest

from TestEnv import TestEnv


def setup_module(module):
    print("setup_module: %s" % module.__name__)
    TestEnv.init()
    TestEnv.a2md_stdargs([TestEnv.A2MD, "-d", TestEnv.STORE_DIR, "-j"])
    TestEnv.a2md_rawargs([TestEnv.A2MD, "-d", TestEnv.STORE_DIR])


def teardown_module(module):
    print("teardown_module: %s" % module.__name__)


class TestRegUpdate:

    NAME1 = "greenbytes2.de"
    NAME2 = "test-100.com"

    def setup_method(self, method):
        print("setup_method: %s" % method.__name__)
        TestEnv.clear_store()
        # add managed domains
        domains = [ 
            [self.NAME1, "www.greenbytes2.de", "mail.greenbytes2.de"],
            [self.NAME2, "test-101.com", "test-102.com"]
        ]
        for dns in domains:
            TestEnv.a2md(["-a", TestEnv.ACME_URL, "add"] + dns)

    def teardown_method(self, method):
        print("teardown_method: %s" % method.__name__)

    # test case: update domains
    def test_110_000(self):
        dns = ["foo.de", "bar.de"]
        output1 = TestEnv.a2md(["-vvvv", "update", self.NAME1, "domains"] + dns)['jout']['output']
        assert len(output1) == 1
        TestEnv.check_json_contains(output1[0], {
            "name": self.NAME1,
            "domains": dns,
            "contacts": [],
            "ca": {
                "url": TestEnv.ACME_URL,
                "proto": "ACME"
            },
            "state": TestEnv.MD_S_INCOMPLETE
        })
        assert TestEnv.a2md(["list"])['jout']['output'][0] == output1[0]

    # test case: remove all domains
    def test_110_001(self):
        assert TestEnv.a2md(["update", self.NAME1, "domains"])['rv'] == 1

    # test case: update domains with invalid DNS
    @pytest.mark.parametrize("invalid_dns", [
        "tld", "white sp.ace", "invalid.*.wildcard.com", "k\xc3ller.idn.com"
    ])
    def test_110_002(self, invalid_dns):
        assert TestEnv.a2md(["update", self.NAME1, "domains", invalid_dns])['rv'] == 1

    # test case: update domains with overlapping DNS list
    def test_110_003(self):
        dns = [self.NAME1, self.NAME2]
        assert TestEnv.a2md(["update", self.NAME1, "domains"] + dns)['rv'] == 1

    # test case: update with subdomains
    def test_110_004(self):
        dns = ["test-foo.com", "sub.test-foo.com"]
        md = TestEnv.a2md(["update", self.NAME1, "domains"] + dns)['jout']['output'][0]
        assert md['name'] == self.NAME1
        assert md['domains'] == dns

    # test case: update domains with duplicates
    def test_110_005(self):
        dns = [self.NAME1, self.NAME1, self.NAME1]
        md = TestEnv.a2md(["update", self.NAME1, "domains"] + dns)['jout']['output'][0]
        assert md['name'] == self.NAME1
        assert md['domains'] == [self.NAME1]

    # test case: remove domains with punycode
    def test_110_006(self):
        dns = [self.NAME1, "xn--kller-jua.punycode.de"]
        md = TestEnv.a2md(["update", self.NAME1, "domains"] + dns)['jout']['output'][0]
        assert md['name'] == self.NAME1
        assert md['domains'] == dns

    # test case: update non-existing managed domain
    def test_110_007(self):
        assert TestEnv.a2md(["update", "test-foo.com", "domains", "test-foo.com"])['rv'] == 1

    # test case: update domains with DNS wildcard
    @pytest.mark.parametrize("wild_dns", [
        "*.wildcard.com"
    ])
    def test_110_008(self, wild_dns):
        assert TestEnv.a2md(["update", self.NAME1, "domains", wild_dns])['rv'] == 0
    
    # --------- update ca ---------

    # test case: update CA URL
    def test_110_100(self):
        url = "http://localhost.com:9999"
        output = TestEnv.a2md(["update", self.NAME1, "ca", url])['jout']['output']
        assert len(output) == 1
        TestEnv.check_json_contains(output[0], {
            "name": self.NAME1,
            "domains": [self.NAME1, "www.greenbytes2.de", "mail.greenbytes2.de"],
            "contacts": [],
            "ca": {
                "url": url,
                "proto": "ACME"
            },
            "state": TestEnv.MD_S_INCOMPLETE
        })

    # test case: update CA with invalid URL
    @pytest.mark.parametrize("invalid_url", [
        "no.schema/path", "http://white space/path", "http://bad.port:-1/path"
    ])
    def test_110_101(self, invalid_url):
        assert TestEnv.a2md(["update", self.NAME1, "ca", invalid_url])['rv'] == 1

    # test case: update ca protocol
    def test_110_102(self):
        md = TestEnv.a2md(["update", self.NAME1, "ca", TestEnv.ACME_URL, "FOO"])['jout']['output'][0]
        TestEnv.check_json_contains(md['ca'], {
            "url": TestEnv.ACME_URL,
            "proto": "FOO"
        })
        assert md['state'] == 1

    # test case: update account ID
    def test_110_200(self):
        acc_id = "test.account.id"
        output = TestEnv.a2md(["update", self.NAME1, "account", acc_id])['jout']['output']
        assert len(output) == 1
        TestEnv.check_json_contains(output[0], {
            "name": self.NAME1,
            "domains": [self.NAME1, "www.greenbytes2.de", "mail.greenbytes2.de"],
            "contacts": [],
            "ca": {
                "account": acc_id,
                "url": TestEnv.ACME_URL,
                "proto": "ACME"
            },
            "state": TestEnv.MD_S_INCOMPLETE
        })

    # test case: remove account ID
    def test_110_201(self):
        assert TestEnv.a2md(["update", self.NAME1, "account", "test.account.id"])['rv'] == 0
        md = TestEnv.a2md(["update", self.NAME1, "account"])['jout']['output'][0]
        TestEnv.check_json_contains(md['ca'], {
            "url": TestEnv.ACME_URL,
            "proto": "ACME"
        })
        assert md['state'] == 1

    # test case: change existing account ID
    def test_110_202(self):
        assert TestEnv.a2md(["update", self.NAME1, "account", "test.account.id"])['rv'] == 0
        md = TestEnv.a2md(["update", self.NAME1, "account", "foo.test.com"])['jout']['output'][0]
        TestEnv.check_json_contains(md['ca'], {
            "account": "foo.test.com",
            "url": TestEnv.ACME_URL,
            "proto": "ACME"
        })
        assert md['state'] == 1

    # test case: ignore additional argument
    def test_110_203(self):
        md = TestEnv.a2md(["update", self.NAME1, "account", "test.account.id",
                           "test2.account.id"])['jout']['output'][0]
        TestEnv.check_json_contains(md['ca'], {
            "account": "test.account.id",
            "url": TestEnv.ACME_URL,
            "proto": "ACME"
        })
        assert md['state'] == 1

    # test case: add contact info
    def test_110_300(self):
        mail = "test@greenbytes.de"
        output = TestEnv.a2md(["update", self.NAME1, "contacts", mail])['jout']['output']
        assert len(output) == 1
        TestEnv.check_json_contains(output[0], {
            "name": self.NAME1,
            "domains": [self.NAME1, "www.greenbytes2.de", "mail.greenbytes2.de"],
            "contacts": ["mailto:" + mail],
            "ca": {
                "url": TestEnv.ACME_URL,
                "proto": "ACME"
            },
            "state": TestEnv.MD_S_INCOMPLETE
        })

    # test case: add multiple contact info, preserve order
    def test_110_301(self):
        mail = ["xxx@greenbytes.de", "aaa@greenbytes.de"]
        md = TestEnv.a2md(["update", self.NAME1, "contacts"] + mail)['jout']['output'][0]
        assert md['contacts'] == ["mailto:" + mail[0], "mailto:" + mail[1]]
        assert md['state'] == 1

    # test case: must not remove contact info
    def test_110_302(self):
        assert TestEnv.a2md(["update", self.NAME1, "contacts", "test@greenbytes.de"])['rv'] == 0
        assert TestEnv.a2md(["update", self.NAME1, "contacts"])['rv'] == 1

    # test case: replace existing contact info
    def test_110_303(self):
        assert TestEnv.a2md(["update", self.NAME1, "contacts", "test@greenbytes.de"])['rv'] == 0
        md = TestEnv.a2md(["update", self.NAME1, "contacts", "xxx@greenbytes.de"])['jout']['output'][0]
        assert md['contacts'] == ["mailto:xxx@greenbytes.de"]
        assert md['state'] == 1

    # test case: use invalid mail address
    @pytest.mark.parametrize("invalid_mail", [
        "no.at.char", "with blank@test.com", "missing.host@", "@missing.localpart.de",
        "double..dot@test.com", "double@at@test.com"
    ])
    def test_110_304(self, invalid_mail):
        # SEI: Uhm, es ist nicht sinnvoll, eine komplette verification von
        # https://tools.ietf.org/html/rfc822 zu bauen?
        assert TestEnv.a2md(["update", self.NAME1, "contacts", invalid_mail])['rv'] == 1

    # test case: respect urls as given
    @pytest.mark.parametrize("url", [
        "mailto:test@greenbytes.de", "wrong://schema@test.com"])
    def test_110_305(self, url):
        md = TestEnv.a2md(["update", self.NAME1, "contacts", url])['jout']['output'][0]
        assert md['contacts'] == [url]
        assert md['state'] == 1

    # test case: add tos agreement
    def test_110_400(self):
        output = TestEnv.a2md(["update", self.NAME1, "agreement", TestEnv.ACME_TOS])['jout']['output']
        assert len(output) == 1
        TestEnv.check_json_contains(output[0], {
            "name": self.NAME1,
            "domains": [self.NAME1, "www.greenbytes2.de", "mail.greenbytes2.de"],
            "contacts": [],
            "ca": {
                "url": TestEnv.ACME_URL,
                "proto": "ACME",
                "agreement": TestEnv.ACME_TOS
            },
            "state": TestEnv.MD_S_INCOMPLETE
        })

    # test case: remove tos agreement
    def test_110_402(self):
        assert TestEnv.a2md(["update", self.NAME1, "agreement", TestEnv.ACME_TOS])['rv'] == 0
        md = TestEnv.a2md(["update", self.NAME1, "agreement"])['jout']['output'][0]
        TestEnv.check_json_contains(md['ca'], {
            "url": TestEnv.ACME_URL,
            "proto": "ACME"
        })
        assert md['state'] == 1

    # test case: ignore additional arguments
    def test_110_403(self):
        md = TestEnv.a2md(["update", self.NAME1, "agreement",
                           TestEnv.ACME_TOS, "http://invalid.tos/"])['jout']['output'][0]
        TestEnv.check_json_contains(md['ca'], {
            "url": TestEnv.ACME_URL,
            "proto": "ACME",
            "agreement": TestEnv.ACME_TOS
        })
        assert md['state'] == 1

    # test case: update agreement with invalid URL
    @pytest.mark.parametrize("invalid_url", [
        "no.schema/path", "http://white space/path", "http://bad.port:-1/path"
    ])
    def test_110_404(self, invalid_url):
        assert TestEnv.a2md(["update", self.NAME1, "agreement", invalid_url])['rv'] == 1
